# coding: utf-8
"""
    Baseline.BaseScattering.py

    Copyright (c) 2020-2025, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.gridspec import GridSpec
from molass_legacy.DataStructure.MatrixData import simple_plot_3d
from .Baseline import compute_baseline

TITLE_FONTSIZE = 16

def primary_peak_topx(ecurve):
    return ecurve.primary_peak_i

def old_demo(root, in_folder, show_preview=False):
    import molass_legacy.KekLib.DebugPlot as plt
    from MeasuredData import MeasuredData

    print(in_folder)
    md = MeasuredData(in_folder)

    if show_preview:
        from molass_legacy.Tools.ThreeDimViewer import ThreeDimViewer
        dialog = ThreeDimViewer(root, md)
        dialog.show()

    xr = md.xr
    xr.set_elution_curve()
    i_slice = xr.i_slice
    j_slice = xr.j_slice
    i_index = xr.e_index - i_slice.start

    whole_data = xr.data
    data = whole_data[i_slice,j_slice]
    q = xr.vector[i_slice]

    fig = plt.figure(figsize=(21, 11))
    gs = GridSpec(2,3)
    ax00 = fig.add_subplot(gs[0,0], projection='3d')
    ax01 = fig.add_subplot(gs[0,1])
    ax02 = fig.add_subplot(gs[0,2])
    ax10 = fig.add_subplot(gs[1,0])
    ax11 = fig.add_subplot(gs[1,1])
    ax12 = fig.add_subplot(gs[1,2])

    ax00.set_title("3D Data View", y=1.09, fontsize=TITLE_FONTSIZE)
    simple_plot_3d(ax00, data, x=q)
    z = xr.e_curve.y[j_slice]
    start = 0 if j_slice.start is None else j_slice.start
    topx = primary_peak_topx(xr.e_curve) - start
    print('topx=', topx)

    q1 = xr.vector[i_index]
    size = data.shape[1]
    x = np.ones(size)*q1
    y = np.arange(size)

    yp = np.average(data[:,topx-10:topx+10], axis=1)
    yd = np.average(data[:,0:50], axis=1) - np.average(data[:,-50:], axis=1)
    scale = yp[i_index]/yd[i_index]
    yd_ = yd*scale

    ax00.plot(x, y, z, color='orange')

    baseline = None
    for ax1, ax2, integral in [
            (ax01, ax02, False), 
            (ax11, ax12, True),
            ]:
        if integral:
            title = "LPM Integral Baseline"
        else:
            title = "LPM Linear Baseline"
        b = compute_baseline(z, integral=integral)
        if baseline is None:
            baseline = b
        z_ = z - b

        ax1.set_title(title, fontsize=TITLE_FONTSIZE)
        ax1.plot(z, color='orange')
        ax1.plot(b, ':', color='red')
        ax1.plot(topx, z[topx], 'o', color='yellow')

        ax2.plot(q, np.log10(yp), label='peak top')
        ax2.plot(q, np.log10(yd_), label='base diff')

        C = np.array([z_, b])
        P = np.dot(data, np.linalg.pinv(C))

        y_ = P[:,0]
        scale = yp[i_index]/y_[i_index]
        ax2.plot(q, np.log10(y_*scale), label='MPI peak')

        yb = P[:,1]
        scale = yp[i_index]/yb[i_index]
        ax2.plot(q, np.log10(yb*scale), label='MPI base')

        ax2.legend()

    fig.tight_layout()
    plt.show()

def demo(parent, in_folder, force_integral=False):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Mapping.MapperConstructor import create_mapper

    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_sd(debug=False)
    pre_recog = PreliminaryRecognition(sd)
    sd_ = pre_recog.get_analysis_copy()

    if force_integral:
        from molass_legacy._MOLASS.SerialSettings import set_setting
        set_setting("baseline_manually", 1)
        set_setting("xray_baseline_type", 2)

    mapper = create_mapper(parent, sd_)
    mapped_info = mapper.get_mapped_info()
    end_slices = mapper.x_curve.get_end_slices()
    basescattering_correct(mapped_info, sd_.intensity_array, end_slices=end_slices, debug=True)

def get_baseplane(baseline, end_slices, yb):
    base_hights = [np.average(baseline[s]) for s in end_slices]
    n_baseline = base_hights[0] + (baseline - base_hights[0])/(base_hights[1] - base_hights[0])
    return np.dot(yb[:,np.newaxis], n_baseline[np.newaxis,:])

def basescattering_correct(mapped_info, intensity_array, use_mpi=False, end_slices=None, debug=False):
    x_curve = mapped_info.x_curve
    x_base = mapped_info.x_base

    x = x_curve.x
    y = x_curve.y       # note that y has been already corrected.

    if debug:
        from bisect import bisect_right
        from molass_legacy._MOLASS.SerialSettings import get_xray_picking
        print("end_slices=", end_slices)
        qv = intensity_array[0,:,0]
        picking = get_xray_picking()
        i = bisect_right(qv, picking)
        with plt.Dp():
            fig, ax = plt.subplots()
            init_ey = intensity_array[:,i,1].copy()
            ax.set_title("basescattering_correct (1)")
            ax.plot(x, init_ey, label="init_ey")
            ax.plot(x, y, color='orange', label="x_curve.y")
            for s in end_slices:
                ax.plot(x[s], y[s], 'o', color='yellow')
            ax.plot(x, x_base, ':', color='red', label="mapped_info.x_base")
            ax.legend()
            fig.tight_layout()
            plt.show()

    M = intensity_array[:,:,1].T
    if use_mpi or debug:
        C = np.array([y, x_base])/x_curve.max_y
        P = np.dot(M, np.linalg.pinv(C))
        yb = P[:,1]
        # TODO: scale
        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("basescattering_correct (2)")
                ax.set_yscale("log")
                for k, yp in enumerate(P.T):
                    ax.plot(qv, yp, label="P[:,%d]" % k)
                ax.legend()
                fig.tight_layout()
                plt.show()

    if not use_mpi:
        assert end_slices is not None
        yb_ = np.average(M[:,end_slices[1]], axis=1) - np.average(M[:,end_slices[0]], axis=1)
        if debug:
            print("M.shape=", M.shape, "end_slices=", end_slices)
            pti = x_curve.primary_peak_i
            qv = intensity_array[0,:,0]
            scale1 = y[pti]
            scale2 = x_base[pti]
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("basescattering_correct (3)")
                ax.set_yscale('log')
                yp = M[:,pti]
                ax.plot(qv, yp, label='peak top')
                ax.plot(qv, P[:,0]*scale1, label='MPI corrected')
                ax.plot(qv, yp - yb_, label='Diff corrected')
                ax.plot(qv, yb*scale2, label="MPI base", alpha=0.3)
                ax.plot(qv, yb_, label='Diff base', alpha=0.3)
                ax.legend()
                fig.tight_layout()
                plt.show()
        yb = yb_

    B = get_baseplane(x_base, end_slices, yb)
    intensity_array[:,:,1] -= B.T

    if debug:
        with plt.Dp():
            fig = plt.figure(figsize=(14,7))
            fig.suptitle("basescattering_correct (4)")
            ax1 = fig.add_subplot(121, projection='3d')
            ax2 = fig.add_subplot(122)
            simple_plot_3d(ax1, B)
            ax2.plot(x, init_ey)
            ax2.plot(x, intensity_array[:,i,1])
            ax2.plot(x, y, color='orange')
            fig.tight_layout()
            plt.show()

    return B



# coding: utf-8
"""

    ModeledData.py

    Copyright (c) 2019, SAXS Team, KEK-PF

"""
import numpy as np
from molass_legacy.Models.ElutionCurveModels import emg
from DataModels import GuinierPorod
import molass_legacy.KekLib.DebugPlot as plt
from MatrixData import simple_plot_3d

class ModeledData:
    def __init__(self, ivector, n_elutions, rg_list=[25], d_list=[1], max_height=1, noise=0, h_list=None, mu_list=None, sigma_list=None, tau_list=None):
        assert len(rg_list) == len(d_list)
        self.num_components = len(rg_list)
        self.i = ivector
        self.j = np.arange(n_elutions)
        self.rg_list = rg_list
        self.d_list = d_list
        self.max_height = max_height
        self.num_peaks = len(rg_list)
        self.make_C_matrix(h_list, mu_list, sigma_list, tau_list)
        self.make_P_matrix()

    def make_C_matrix(self, h_list, mu_list, sigma_list, tau_list):
        if mu_list is None:
            self.top_i = np.linspace(self.j[0], self.j[-1], self.num_peaks+2)[1:-1]
        else:
            self.top_i = mu_list
        C_list = []
        self.e_params = []
        for k in range(self.num_peaks):
            h = 1 if h_list is None else h_list[k]
            mu = self.top_i[k]
            sigma = len(self.j)/10 if sigma_list is None else sigma_list[k]
            tau = 0 if tau_list is None else tau_list[k]
            C_list.append(emg(self.j, h=h, mu=mu, sigma=sigma, tau=tau))
            self.e_params.append((h, mu, sigma, tau))
        self.C = np.array(C_list)

    def make_P_matrix(self):
        h = self.max_height
        d = 1
        P_list = []
        for rg, d in zip(self.rg_list, self.d_list):
            gp_model = GuinierPorod(h, rg, d)
            P_list.append(gp_model(self.i))
        self.P = np.array(P_list).T

    def get_data(self, noise=0.01, debug=False):
        M = np.dot(self.P, self.C)
        if noise > 0:
            M_ = M/np.max(M)
            M *= (np.random.normal(0, 1, M.shape)*noise/M_ + 1)

        if debug:
            # fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21,7), subplot_kw={'projection': '3d'})
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            simple_plot_3d(ax, M, x=self.i)
            fig.tight_layout()
            plt.show()

        return M

    def plot_components(self):
        n = self.num_components
        fig = plt.figure(figsize=(21, 6))
        ax1 = fig.add_subplot(131)
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133, projection='3d')
        for i in range(n):
            ax1.plot(self.i, self.P[:,i])
            ax2.plot(self.C[i,:])
        M = self.get_data()
        simple_plot_3d(ax3, M, x=self.i)
        fig.tight_layout()
        plt.show()



"""
    Models.Stochastic.MonoporeUvScaler.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SecTheory.BasicModels import robust_single_pore_pdf

def adjust_to_uv_scales(x, y, uv_x, uv_y_, xr_init_params, rgs, debug=False):
    print("adjust_to_uv_scales: xr_init_params=", xr_init_params, "rgs=", rgs)
    N, T, t0, me, mp, poresize = xr_init_params[0:6]
    xr_scales = xr_init_params[6:]
    rhov = rgs/poresize
    rhov[rhov > 1] = 1
    cy_list = []
    for rho in rhov:
        np_ = (1 - rho)**me
        tp_ = (1 - rho)**mp
        cy = robust_single_pore_pdf(x - t0, N * np_, T * tp_)   # note that the scale is not applied here
        cy_list.append(cy)
    ty = np.sum(cy_list, axis=0)

    def uv_scales_objective(scales, return_cy=False):
        uv_cy_list = []
        for cy, scale in zip(cy_list, scales):
            uv_cy = scale * cy
            uv_cy_list.append(uv_cy)
        uv_ty = np.sum(uv_cy_list, axis=0)
        if return_cy:
            return uv_cy_list, uv_ty
        dev = np.sum((uv_ty - uv_y_)**2)
        return dev

    res = minimize(uv_scales_objective, xr_scales, method="Nelder-Mead")
    uv_cy_list, uv_ty = uv_scales_objective(res.x, return_cy=True)

    if debug:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
            fig.suptitle("adjust_to_uv_scales")
            ax1.set_title("UV")

            ax1.plot(uv_x, uv_y_, color="blue")
            for k, cy in enumerate(uv_cy_list):
                ax1.plot(uv_x, cy, ":", label='component-%d' % k)
            ax1.plot(uv_x, uv_ty, ":", color="red", label='model total')
            ax1.legend()

            ax2.set_title("XR")
            ax2.plot(x, y, color="orange")
            xr_cy_list = []
            for k, (cy, scale) in enumerate(zip(cy_list, xr_scales)):
                xr_cy = scale * cy
                ax2.plot(x, xr_cy, ":", label='component-%d' % k)
                xr_cy_list.append(xr_cy)
            xr_ty = np.sum(xr_cy_list, axis=0)
            ax2.plot(x, xr_ty, ":", color="red", label='model total')
            ax2.legend()

            fig.tight_layout()
            ret = plt.show()
    else:
        ret = True

    if ret:
        return res.x, uv_ty
    else:
        return


"""
    RangeInfo.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF

    Eiher of these infos can be converted to an AnalysisRangeInfo

                                SerialSetting temporaries
                                -----------------------
        RangeEditorInfo    <=>   range_editor_info
        DecompEditorInfo   <=>   decomp_editor_info

             ↓

        AnalysisRangeInfo  <=>   analysis_range_info
"""
import copy

class SuperEditorInfo:
    def get_ranges(self):
        ranges = []
        for prange in self.paired_ranges:
            try:
                range_list = prange.get_fromto_list()
            except:
                range_list = prange[1:]
            ranges.append(range_list)
        return ranges

    def get_ignorable_flags(self):
        return self.ignorable_flags

    def __repr__(self):
        return "%s(%s,%s)" % (self.__class__.__qualname__,
                              self.paired_ranges,
                              repr(self.ignorable_flags),   # this will stringify, e.g., as 'array([False, ..., ])'
                              )

class RangeEditorInfo(SuperEditorInfo):
    def __init__(self, paired_ranges, ignorable_flags):
        self.paired_ranges = copy.deepcopy(paired_ranges)
        self.ignorable_flags = copy.deepcopy(ignorable_flags)

    def update(self, default_ranges):
        pass

class DecompEditorInfo(SuperEditorInfo):
    def __init__(self, paired_ranges, ignorable_flags):
        self.paired_ranges = copy.deepcopy(paired_ranges)
        self.ignorable_flags = copy.deepcopy(ignorable_flags)

    def update(self, control_info):
        pass

def shift_editor_ranges(j0, editor_ranges):
    ret_range_list = []
    for prange in editor_ranges:
        prange_ = []
        for f, t in prange:
            prange_.append([j0+f, j0+t])
        ret_range_list.append(prange_)

    return ret_range_list



"""
    ManualAdjuster.py

    Copyright (c) 2019-2023, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
try:
    import molass_legacy.KekLib.CustomMessageBox as MessageBox
except:
    import OurMessageBox as MessageBox
from molass_legacy.Elution.CurveUtils import simple_plot
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

def single_float(f):
    return float('%g' % f)

class ManualAdjuster(Dialog):
    def __init__(self, parent, sd, mapper):
        self.a_curve = sd.absorbance.a_curve
        self.x_curve = sd.xray_curve
        A, B = mapper.map_params
        self.init_A = A
        self.init_B = B

        time_shift = get_setting('manual_time_shift')
        time_scale = get_setting('manual_time_scale')
        if time_shift is not None:
            A = 1/time_scale
            B = -time_shift/time_scale

        self.A = A
        self.B = B
        self.applied = False

        Dialog.__init__(self, parent, "Manual Syncronizer", visible=False)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        iframe = Tk.Frame(body_frame)
        iframe.pack(padx=20, pady=10)

        fig_frame = Tk.Frame(iframe)
        fig_frame.pack()
        lower_frame = Tk.Frame(iframe)
        lower_frame.pack(fill=Tk.X)

        fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(21, 7))
        self.mpl_canvas = FigureCanvasTkAgg(fig, fig_frame)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.fig = fig
        fig.suptitle("Manual Adjustment for " + get_setting('in_folder'), fontsize=20)
        self.axes = axes

        self.draw_figures()

        tool_frame = Tk.Frame(lower_frame)
        tool_frame.pack(side=Tk.LEFT)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tool_frame)
        self.toolbar.update()

        manip_frame = Tk.Frame(lower_frame)
        manip_frame.pack(side=Tk.RIGHT, padx=100)

        self.build_manip_panel(manip_frame)

    def draw_figures(self):
        for ax in self.axes:
            ax.cla()

        a_curve = self.a_curve
        x_curve = self.x_curve 

        ax1, ax2, ax3 = self.axes
        ax1.set_title("UV data", fontsize=16)
        ax2.set_title("Xray data", fontsize=16)
        ax3.set_title("UV/Xray overlay (Xray intensity scale)", fontsize=16)

        simple_plot(ax1, a_curve, color='blue')
        simple_plot(ax2, x_curve, color='orange')
        simple_plot(ax3, x_curve, color='orange', legend=False)

        x = x_curve.x
        j = self.A*x + self.B
        scale = x_curve.max_y/a_curve.max_y
        y = a_curve.spline(j)*scale
        ax3.plot(x, y, color='blue', label='mapped data')

        ax1.set_xlim(ax1.get_xlim())
        ax1.set_ylim(ax1.get_ylim())
        ax1.plot( ax1.get_xlim(), [0, 0], ':', color='red' )

        xlim_list = []
        ylim_list = []
        for ax in [ax2, ax3]:
            xmin, xmax = ax.get_xlim()
            xlim_list.append((xmin, xmax))
            ymin, ymax = ax.get_ylim()
            ylim_list.append((ymin, ymax))

        xlim_array = np.array(xlim_list)
        ylim_array = np.array(ylim_list)

        xmin = np.min(xlim_array[:,0])
        xmax = np.max(xlim_array[:,1])
        ymin = np.min(ylim_array[:,0])
        ymax = np.min(ylim_array[:,1])

        for ax in [ax2, ax3]:
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)
            ax.plot( [xmin, xmax], [0, 0], ':', color='red' )

        ax3.legend()

        self.fig.tight_layout()
        self.fig.subplots_adjust(top=0.85)
        self.mpl_canvas.draw()

    def build_manip_panel(self, manip_frame):
        shift_label = Tk.Label(manip_frame, text="time shift")
        shift_label.grid(row=0, column=0)

        max_shift = len(self.x_curve.x)*2
        self.time_shift = Tk.DoubleVar()
        self.time_shift.set(single_float(-self.B/self.A))
        shift_spinbox = Tk.Spinbox(manip_frame, textvariable=self.time_shift,
                                            from_=-max_shift, to=max_shift, increment=0.1,
                                            justify=Tk.CENTER, width=8)
        shift_spinbox.grid(row=0, column=1)

        space = Tk.Label(manip_frame, width=4)
        space.grid(row=0, column=2)

        scale_label = Tk.Label(manip_frame, text="time scale")
        scale_label.grid(row=0, column=3)

        min_scale = 0.1
        max_scale = 2
        self.time_scale = Tk.DoubleVar()
        self.time_scale.set(single_float(1/self.A))
        scale_spinbox = Tk.Spinbox(manip_frame, textvariable=self.time_scale,
                                            from_=min_scale, to=max_scale, increment=0.001,
                                            justify=Tk.CENTER, width=8)
        scale_spinbox.grid(row=0, column=4)

        self.time_shift.trace('w', self.time_params_tracer)
        self.time_scale.trace('w', self.time_params_tracer)

        space = Tk.Label(manip_frame, width=4)
        space.grid(row=0, column=5)

        reset_btn = Tk.Button(manip_frame, text="Reset", command=self.reset)
        reset_btn.grid(row=0, column=6)

    def show(self):
        self._show()

    def time_params_tracer(self, *args):
        try:
            time_shift = self.time_shift.get()
            time_scale = self.time_scale.get()
            self.A = 1/time_scale               # need to cover zero divide
            self.B = -time_shift/time_scale
        except:
            return

        self.draw_figures()

    def reset_to_the_init_params(self):
        A, B = self.init_A, self.init_B

        A_ = single_float(1/A)      # so as to look like single precision
        B_ = single_float(-B/A)     # so as to look like single precision

        print('B_=', B_, 'A_=', A_)
        self.time_shift.set(B_)
        self.time_scale.set(A_)

    def reset(self):
        self.reset_to_the_init_params()
        self.draw_figures()

    def apply(self):
        self.applied = True

        time_shift = self.time_shift.get()
        time_scale = self.time_scale.get()

        set_setting('manual_time_shift', time_shift)
        set_setting('manual_time_scale', time_scale)



# coding: utf-8
"""
    OurPipApi.py

    Copyright (c) 2018-2019, Masatsuyo Takahashi, KEK-PF
"""

import re
import fnmatch
import os
import subprocess
from packaging import version
from pip_api import installed_distributions
# from pip_api._call import call
from Console import reset_text_color

def decode_robust(b):
    try:
        s = b.decode('utf-8', errors='ignore')
    except Exception as exc:
        s = str(exc)
    return s

def call(*args):
    python_location = os.environ.get('PIPAPI_PYTHON_LOCATION', 'python')
    result = subprocess.check_output(
        [python_location, '-m', 'pip'] + list(args)
    )
    return decode_robust(result)

shown_line_re = re.compile(r'^(\S+):\s*(.*)$')
searched_line_re = re.compile(r'^(\S+)')
installed_line_re = re.compile(r'^\s+INSTALLED:\s+(\S+)\s*(\(latest\))?')
latest_line_re = re.compile(r'^\s+LATEST:\s+(.+)')

def my_print(*args):
    global _logger
    if _logger is None:
        print(*args)
    else:
        msg = ' '.join( [str(v) for v in args] )
        _logger.info(msg)

def update_all_packages(restrict=None, exec_install=False, logger=None):
    global _logger
    _logger = logger
    if restrict is None:
        restrict_re = None
    else:
        restrict_re = re.compile(fnmatch.translate(restrict))

    for k, v in installed_distributions().items():
        if k[0] == '-':
            my_print(k, "seems to have been erroneously installed.")
            continue

        if restrict_re is None:
            pass
        else:
            if restrict_re.match(k):
                pass
            else:
                continue
        reset_text_color()
        my_print(k)
        shown = call('show', k)
        # my_print(shown)
        for line in shown.split('\r\n'):
            m = shown_line_re.match(line)
            if m:
                key = m.group(1)
                value = m.group(2)
                if key == 'Name':
                    assert k == value
                    # my_print("'%s'" % value)
                elif key == 'Version':
                    my_print(k, value)
        try:
            searched = call('search', k)
        except Exception as exc:
            my_print(exc)
            my_print('pip search failed for', k)
            searched = ''
        # my_print(searched)
        this_line_no = None
        verified = False
        for n, line in enumerate(searched.split('\r\n')):
            m = searched_line_re.search(line)
            if m:
                name = m.group(1)
                if name == k:
                    # my_print(line)
                    this_line_no = n
            if this_line_no is not None and n > this_line_no and n < this_line_no + 3:
                my_print(line)
                m = installed_line_re.match(line)
                if m:
                    installed_version = m.group(1)
                    installed_latest = m.group(2)
                    # my_print(installed_version, installed_latest)
                    if installed_latest is not None:
                        verified = True
                        break
                m = latest_line_re.match(line)
                if m:
                    latest_version = m.group(1)
                    my_print(installed_version, latest_version)
                    if version.parse(installed_version) < version.parse(latest_version):
                        my_print('need update')
                        if exec_install:
                            my_print('pip install %s -U' % k)
                            try:
                                install_log = call('install', k, '-U')
                                my_print(install_log)
                                if install_log.lower().find('error') > 0:
                                    exit()
                            except Exception as exc:
                                my_print(exc)
                                my_print('pip install failed for', k)
                    else:
                        my_print('no need')
                    verified = True
                    break

        # assert verified
        if not verified:
            my_print( k, 'is not verified. need to update manually.' )



"""
    DriftAnalyzer.py

    Copyright (c) 2018-2023, Masatsuyo Takahashi, KEK-PF
"""
import copy
import numpy as np
from scipy                      import stats
from IntensityData              import IntensityData
from GuinierPorodFit            import GuinierPorodFit
from molass_legacy.Mapping.MapperConstructor import create_mapper
from ScatteringBaseCorrector    import ScatteringBaseCorrector
from DriftLinearModel           import DriftLinearModel

USE_GUINIER_INTERVAL_RATIO  = False
GUINIER_FIT_START_RATIO     = 0.005
GUINIER_FIT_STOP_RATIO      = 0.1

def create_mapper_for_drift_analysis( parent, sd, opt_params=None ):

    if opt_params is None:
        opt_params = {}
        opt_params[ 'uv_baseline_opt' ]         = 1
        opt_params[ 'uv_baseline_type' ]        = 1
        opt_params[ 'uv_baseline_adjust' ]      = 0
        opt_params[ 'xray_baseline_opt' ]       = 1
        opt_params[ 'xray_baseline_type' ]      = 1
        opt_params[ 'xray_baseline_adjust' ]    = 0
        opt_params[ 'dev_allow_ratio' ]         = 0.5

    mapper = create_mapper( parent, sd, opt_params=opt_params )
    return mapper

def get_baselines( sd, mapper, indeces ):

    data_copy = copy.deepcopy( sd.intensity_array )

    params  = mapper.opt_params

    corrector = ScatteringBaseCorrector(
                            sd.jvector,
                            sd.qvector,
                            data_copy,
                            curve=mapper.sd_xray_curve,
                            affine_info=mapper.get_affine_info(),
                            inty_curve_y=mapper.x_curve.y,
                            baseline_opt=params['xray_baseline_opt'],
                            baseline_degree=params['xray_baseline_const_opt'] + 1,
                            need_adjustment=params['xray_baseline_adjust'] == 1,
                            )
    baseline_list = []
    for i in indeces:
        baseline = corrector.correct_a_single_q_plane( i, return_baseline=True )
        baseline_list.append( baseline )

    return baseline_list

class DriftAnalyzer:
    def __init__( self, sd, mapper ):
        self.sd     = sd
        self.mapper = mapper
        self.ecurve = mapper.x_curve
        self.data   = sd.intensity_array

    def solve( self ):
        sd  = self.sd
        qsize   = self.data.shape[1]
        if USE_GUINIER_INTERVAL_RATIO:
            start   = int( qsize * GUINIER_FIT_START_RATIO )
            stop    = int( qsize * GUINIER_FIT_STOP_RATIO )
            print( 'start, stop=', (start, stop) )
        else:
            start   = 5
            stop    = 100
        indeces = np.arange( start, stop, 1, dtype=int )

        baseline_list = get_baselines( sd, self.mapper, indeces )
        model   = DriftLinearModel( sd.qvector, len(sd.jvector) )
        model.fit( indeces, baseline_list )
        print( 'params=', model.params )

        gp_params = []
        for k, info in enumerate(self.ecurve.peak_info):
            print( k, info )
            p   = int( info[1]+0.5 )
            intensity   = IntensityData( self.data[p,:,:], add_smoother=True )
            gp = GuinierPorodFit( intensity )
            gp.fit()
            print( gp.I0, gp.Rg, gp.d )
            gp_params.append( [ gp.I0, gp.Rg, gp.d ] )

        return model.params, gp_params



"""
    Theory.RoigSolvas2019.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from .JsPedersen1997 import SolidEllipsoidRev, TriAxialEllipsoid

def ellipsoid_approx_intensities(q, Rg2, Af2):
    return 1 - Rg2/3 * q**2 + 1/21*(Rg2**2 + 1/5*Af2)*q**4

class Anisotropy:
    def __init__(self):
        pass

    def fit(self, q, Ic, initRg):
        initR = np.asrt(5/3)*initRg

    def fit_approx(self, q, Ic, initRg):

        def obj_func(p):
            Rg2, Af2 = p
            return np.sum((Ic - ellipsoid_approx_intensities(q, Rg2, Af2))**2)

        # method = 'SLSQP'
        # method = 'BFGS'
        # method = 'L-BFGS-B'
        # method = 'CG'
        initRg2 = initRg**2
        # res = minimize(obj_func, (initRg2, 0), method=method, bounds=((0, initRg2*2), (0, 1)))
        res = minimize(obj_func, (initRg2, 0))
        self.C = 1
        self.Rg2, self.Af2 = res.x
        return res.x

    def approx_intensities(self, q):
        return self.C * ellipsoid_approx_intensities(q, self.Rg2, self.Af2)

def demo(root, sd, pno=0, debug=True):
    from bisect import bisect_right
    from .SynthesizedLRF import synthesized_lrf_spike as synthesized_lrf
    from SvdDenoise import get_denoised_data
    from .Rg import compute_corrected_Rg, compute_Rg
    from .SolidSphere import get_boundary_params_simple, SolidSphere
    from .PbMoore1980 import ift_intensities
    from DataUtils import get_in_folder
    from .JsPedersen1997 import P6

    M, E, qv, ecurve = sd.get_xr_data_separate_ly()

    range_ = ecurve.get_ranges_by_ratio(0.5)[pno]
    f = range_[0]
    p = range_[1]
    t = range_[2]

    eslice = slice(f,t)
    x = ecurve.x
    y = ecurve.y

    M0 = M[:,eslice]

    c_ = y[eslice]
    M2 = get_denoised_data(M0, rank=2)
    C2 = np.array([c_, c_**2])
    P2 = M2 @ np.linalg.pinv(C2)

    Rg, gf, gt, _ = compute_corrected_Rg(sd, ecurve, pno, qv, M0, E[:,eslice])
    b1, b2, k = get_boundary_params_simple(Rg)
    P = synthesized_lrf(qv, M0, c_, M2, P2, boundary=b1, k=k)

    spt = P[:,0]
    # qc = np.hstack([[0], qv])
    qc = qv
    qc, Ic, Icerr, D = ift_intensities(qv, spt, E[:,p], qc=qc)
    scale = 1/Ic[0]
    spt_ = spt*scale
    Ic_ = Ic*scale

    sphere = SolidSphere()
    sphere.fit(qc, Ic_, Rg)
    y1 = sphere.intensity(qv)

    ellips = Anisotropy()
    i1 = bisect_right(qc, b1/2)
    ellips.fit_approx(qc[0:i1], Ic_[0:i1], Rg)
    y2 = ellips.approx_intensities(qv)
    print("Af2=%.2g" % ellips.Af2)
    af = np.sqrt(ellips.Af2)
    print("af=%.2g" % af)

    a, b, c = 3, 4, 5
    rg6 = np.sqrt((a**2 + b**2 + c**2)/5)
    scale = Rg/rg6

    p6 = P6(qc, a*scale, b*scale, c*scale)

    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(16,7))
    fig.suptitle("Anisotropy Factor Demo based on Roig-Solvas, 2019", fontsize=24)
    in_folder = get_in_folder()
    ax1.set_title(in_folder, fontsize=20)
    ax1.set_yscale('log')
    ax1.plot(qv, spt_, label='Syn LRF', color='C1')
    ax1.plot(qc, Ic_, label='Syn LRF IFT', color='C2')
    ax1.plot(qc, y1, label='Solid Sphere', color='yellow')
    ax1.plot(qc, y2, label='Fitting (Roig-Solvas, 2019)', color='cyan')

    ax1.text(0.15, 10e-2, "$A_F=%.2g$" % af, fontsize=50, alpha=0.3)

    ax1.legend(fontsize=16)

    i1 = bisect_right(qc, b1/2)
    ellips = Anisotropy()
    ellips.fit_approx(qc[0:i1], p6[0:i1], Rg)
    y2 = ellips.approx_intensities(qv)
    print("Af2=%.2g" % ellips.Af2)
    af = np.sqrt(ellips.Af2)
    print("af=%.2g" % af)

    ax2.set_yscale('log')
    ax2.set_title("Ellissoid(%.2g, %.2g, %.2g)" % (a*scale, b*scale, c*scale), fontsize=20)
    ax2.plot(qc, p6, label="Ellipsoid", color='green')
    ax2.plot(qc, y1, label='Solid Sphere', color='yellow')
    ax2.plot(qc, y2, label='Fitting (Roig-Solvas, 2019)', color='cyan')
    ax2.text(0.15, 10e-2, "$A_F=%.2g$" % af, fontsize=50, alpha=0.3)

    ax2.legend(fontsize=16)

    fig.tight_layout()
    fig.subplots_adjust(top=0.85)
    plt.show()



# coding: utf-8
"""
    SimpleUnfolding.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import logging
import numpy as np
from scipy import optimize
from lmfit import Parameters, minimize

MAX_ITERATION_FMIN_CG = 10000

"""
constants borrowed from PyFolding.constants.py
"""
# constants
IDEAL_GAS_CONSTANT_KCAL = 1.987204118E-3
TEMPERATURE_CELSIUS = 25.

# calculated constants
ZERO_KELVIN = 273.15
TEMPERATURE_KELVIN = ZERO_KELVIN + TEMPERATURE_CELSIUS
RT = IDEAL_GAS_CONSTANT_KCAL * TEMPERATURE_KELVIN
# print('RT=', RT)    # 0.5924849

def proportion_folded(G, m, x):
    return 1/(1 + np.exp(-(G-m*x)/RT))

class SimpleUnfolding:
    def __init__(self, params=None):
        self.logger = logging.getLogger(__name__)
        if params is None:
            params = Parameters()
            params.add('G', value=4)
            params.add('m', value=1)
            params.add('af', value=0.5)
            params.add('bf', value=-0.01)
            params.add('au', value=0.3)
            params.add('bu', value=-0.02)
        self.params = params

    def compute_Pf(self, params, x):
        if params is None:
            params = self.params
        G = params['G']
        m = params['m']
        pf = proportion_folded(G, m, x)
        return pf

    def compute_yf(self, params, pf, x):
        if params is None:
            params = self.params

        af = params['af']
        bf = params['bf']
        yf = (af + bf*x)*pf
        return yf

    def compute_yu(self, params, pf, x):
        if params is None:
            params = self.params

        au = params['au']
        bu = params['bu']
        pu = (1-pf)
        yu = (au + bu*x)*pu
        return yu

    def fit(self, x, y, G_init=1):
        # use UnfoldingModel to get init pramas
        mx = (x[0] + x[-1])/2
        init_params = Parameters()
        init_params.add('G', value=G_init, min=0, max=100)
        m_init = G_init/mx
        init_params.add('m', value=m_init, min=0, max=100)
        init_params.add('af', value=y[0], min=0, max=10)
        init_params.add('bf', value=0, min=-1, max=1)
        init_params.add('au', value=y[-1], min=0, max=10)
        init_params.add('bu', value=0, min=-1, max=1)
        # print('init params=', [(k, p.value) for k, p in init_params.items()])

        def obj_func(p):
            pf = self.compute_Pf(p, x)
            yf = self.compute_yf(p, pf, x)
            yu = self.compute_yu(p, pf, x)
            return yf+yu - y

        min_error = None
        opt_method = None
        opt_result = None
        # for method in ['leastsq', 'least_squares', 'cg', 'tnc', 'dual_annealing']:
        for method in ['leastsq']:
            res = minimize(obj_func, init_params, args=(), method=method)
            error = self.compute_error(res.params)
            if min_error is None or error < min_error:
                min_error = error
                opt_method = method
                opt_result = res

        # print('opt_method=', opt_method)
        G = opt_result.params['G']
        m = opt_result.params['m']
        self.logger.info("initial parameters (G=%.3g, m=%.3g) have been optimized to (G=%.3g, m=%.3g).",
                            G_init, m_init, G, m)
        return opt_result

    def compute_error(self, params):
        error = 0
        for p0, p1 in zip(self.params.items(), params.items()):
            error += (p0[1].value - p1[1].value)**2
        return error

    def make_data(self, dependency, x):
        pf = self.compute_Pf(None, x)
        yf = self.compute_yf(None, pf, x)
        yu = self.compute_yu(None, pf, x)
        return [yf, yu]

    def make_conc_vector(self, x):
        yf, yu = self.make_data(1, x)
        return yf + yu



"""
    CurveFeatures.py

        Optimization of mapping between UV-absorbance and Xray-scattering

    Copyright (c) 2019-2025, SAXS Team, KEK-PF
"""
import numpy as np
import logging
from scipy.interpolate import UnivariateSpline, LSQUnivariateSpline

PEAK_TOP_WEIGHT     = 1.0
MAJOR_POINT_WEIGHT  = 0.5      # MAJOR_POINT_WEIGHT==OTHER_POINT_WEIGHT for 20190305_2
OTHER_POINT_WEIGHT  = 0.5
FOOT_POINT_WEIGHT   = 0.8

POSITIVE_DERIV_ADOPT_RATIO  = 0.1
NEGATIVE_DERIV_ADOPT_RATIO  = 0.1
VALUE_ADOPT_RATIO           = 0.03
ADD_FOOT_POINTS     = False
FOOT_LEVEL_RATIO    = 0.1
FOOT_ADD_WIDTH      = 10

# TODO: remove duplication of this code
def rotate( th, wx, wy ):
    k = -1 if th > 0 else 0
    cx  = wx[k]
    cy  = wy[k]
    wx_ = wx - cx
    wy_ = wy - cy
    c   = np.cos( th )
    s   = np.sin( th )
    return cx + ( wx_*c - wy_* s ), cy + ( wx_*s + wy_* c )

class CurveFeatures:
    def __init__(self, curve, data_label=""):
        self.logger = logging.getLogger(__name__)
        self.curve = curve

        x = curve.x

        d2y = curve.d2y
        d2_roots = curve.d2_roots
        max_d1 = np.max(curve.d1y)
        min_d1 = np.min(curve.d1y)
        roots = []
        for r in d2_roots:
            ratio = curve.spline(r)/curve.max_y
            if ratio < VALUE_ADOPT_RATIO:
               continue

            d1y_r = curve.d1(r)
            if d1y_r > 0:
                if d1y_r/max_d1 > POSITIVE_DERIV_ADOPT_RATIO:
                    roots.append(r)
            else:
                if abs(d1y_r/min_d1) > NEGATIVE_DERIV_ADOPT_RATIO:
                    roots.append(r)

        if ADD_FOOT_POINTS:
            if len(curve.peak_info) == 1 and curve.has_few_points():
                foot_level = FOOT_LEVEL_RATIO * curve.max_y
                shifted_spline = LSQUnivariateSpline(x, curve.y_for_spline-foot_level, curve.knots[1:-1], ext=3)
                foots = shifted_spline.roots()
                left, top, right = curve.peak_info[0]
                for f in foots:
                    if left - FOOT_ADD_WIDTH < f and f < right + FOOT_ADD_WIDTH:
                        roots.append(f)

        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy._MOLASS.SerialSettings import get_setting
            in_folder = get_setting('in_folder')
            fig = plt.figure(figsize=(8,6))
            ax = fig.gca()
            d = 0.01 if in_folder.find('OA_Ald') >= 0 else 0.02
            yticks = np.arange(0, 3, 0.1) if data_label.find('UV') >= 0 else np.arange(0, 0.3, d)
            ax.yaxis.set_ticks(yticks)
            ax.tick_params(labelsize = 24)
            peak_top = False
            if peak_top:
                point_type = 'Peak Top Points'
            else:
                axt = ax.twinx()
                # axt.tick_params(labelsize = 24)
                axt.set_axis_off()

                point_type = 'Inflection Points'
            ax.set_title('%s %s' % (data_label, point_type) , fontsize=40)
            # ax.plot(x, curve.y, 'o', markersize=3)
            ax.plot(x, curve.y, linewidth=3)
            if peak_top:
                for info in curve.peak_info:
                    top_x = info[1]
                    ax.plot(top_x, curve.y[top_x], 'o', color='red', markersize=10)
            else:
                axt.plot(x, curve.d1y, ':', linewidth=3)
                for r in roots:
                    axt.plot(r, curve.d1(r), 'o', color='red', markersize=10)
                    ax.plot(r, curve.spline(r), 'o', color='yellow', markersize=10)
            fig.tight_layout()
            plt.show()

        if False:
            if len(curve.peak_info) == 1 and len(roots) == 2:
                if curve.half_x is not None and len(curve.half_x) == 2:
                    roots = list(curve.half_x)
                    self.logger.info("inflection points have been replaced with half width points " + str(roots))

        self.roots = roots
        self.weights = np.ones(len(roots))*OTHER_POINT_WEIGHT
        if len(curve.peak_info) == 1:
            left, topx, right = curve.peak_info[0]
            width = right - left
            for k, r in enumerate(roots):
                dist_ratio = abs(r - topx)/width
                # print([k], "dist_ratio=", dist_ratio)
                if dist_ratio > 0.5:
                    self.weights[k] = MAJOR_POINT_WEIGHT

    def get_points(self, adopted_peak_info=None, add_feet=False):
        # add_feet=True made 20181203 worse. why?

        if adopted_peak_info is None:
            adopted_peak_info = self.curve.peak_info
        peak_rec_list = [[ info[1], 1.0 ] for info in adopted_peak_info]
        root_rec_list = [[r, w] for r, w in zip(self.roots, self.weights)]
        if add_feet:
            foot_rec_list = [[f, FOOT_POINT_WEIGHT] for f in self.get_foot_points(adopted_peak_info)]
        else:
            foot_rec_list = []
        records = peak_rec_list + root_rec_list + foot_rec_list
        records_ = sorted(records)
        # print('records_=', records_)
        points = np.array([rec[0] for rec in records_])
        weights = np.array([rec[1] for rec in records_])
        return points, weights

    def get_foot_points(self, adopted_peak_info=None):
        num_peaks = len(adopted_peak_info)

        size = len(self.curve.x)
        start = 0
        prev_top_i = None
        foot_points = []

        for k, prec in enumerate(adopted_peak_info):
            top_i = prec[1]

            left_start = start
            left_stop = top_i

            try:
                left_foot = self.find_foot(-1, left_start, left_stop)
                foot_points.append(left_foot)
            except AssertionError:
                pass

            right_start = top_i
            right_stop = size if k == num_peaks - 1 else self.get_valley_bottom(top_i, adopted_peak_info[k+1][1])

            try:
                right_foot = self.find_foot(+1, right_start, right_stop)
                foot_points.append(right_foot)
            except AssertionError:
                pass

            prev_top_i = top_i
            start = right_stop

        return foot_points

    def get_valley_bottom(self, top_i, next_opt_i):
        n = np.argmin(self.curve.y[top_i:next_opt_i])
        return top_i + n

    def find_foot(self, sign, start, stop):

        curve = self.curve

        slice_ = slice(start, stop)

        x = curve.x[slice_]
        y = curve.y[slice_]

        x1, x2 = x[[0, -1]]
        y1, y2 = y[[0, -1]]

        xscale = 1/abs(x2 - x1)
        yscale = 1/abs(y2 - y1)

        x_ = x * xscale
        y_ = y * yscale

        angle = sign* np.pi/4

        rx, ry = rotate(angle, x_, y_)

        pnn = np.argpartition(ry, 5)
        n = int(np.average(pnn[0:5]) + 0.5)

        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.Elution.CurveUtils import simple_plot
            plt.push()
            fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14, 7))
            ax1, ax2 = axes

            fig.suptitle("find_foot ")
            simple_plot(ax1, curve)
            x_ = curve.x[start:stop]
            y_ = curve.y[start:stop]
            ax2.plot(rx, ry)
            ax2.plot(rx[n], ry[n], 'o', color='yellow')
            fig.tight_layout()
            fig.subplots_adjust(top=0.9)
            plt.show()
            plt.pop()

        return start + n



"""
    StructureFactorBounds.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting

class StructureFactorBounds:
    def __init__(self, qv, lrf_info, gk_info):
        self.qv = qv
        rg = np.mean(gk_info.rgs)
        self.R = np.sqrt(5/3)*rg            # hard sphere raidius
        L = self.estimate_L(lrf_info)
        bound = 1/(self.qv*self.R*L)**2
        self.zv = np.zeros(len(qv))
        self.bounds = (-bound, bound)

    def get_bounds(self):
        return self.bounds

    def estimate_L(self, lrf_info):
        Pxr, Cxr, Puv, Cuv, mapped_UvD = lrf_info.matrices
        cv = np.max(Cxr[:-1], axis=1)
        n = np.argmax(cv)
        print("cv=", cv, "n=", n)
        self.n = n
        self.c = cv[n]
        self.eoii_pv = Pxr[-1]
        sf_bound_ratio = get_setting("sf_bound_ratio")
        if sf_bound_ratio < 0:
            # estimate any way
            sf_bound_ratio = 1      # temporarily estimated value
        return sf_bound_ratio

    def compute_penalty(self, Pxr, debug=False):
        penalty = 0
        bq = Pxr[:,-1]
        if debug:
            bq_bounds_list = []
        for aq in Pxr.T[:-1]:
            aq_ = np.abs(aq)
            penalty = np.mean(np.min([self.zv, bq - self.bounds[0]*aq_], axis=0)**2) + np.mean(np.max([self.zv, bq - self.bounds[1]*aq_], axis=0)**2)
            if debug:
                bq_bounds_list.append(self.bounds[0]*aq_)
                bq_bounds_list.append(self.bounds[1]*aq_)

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            qv = self.qv
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.plot(qv, bq)
                for bv in bq_bounds_list:
                    ax.plot(qv, bv)
                plt.show()

        return penalty

    def compute_bounded_bq(self, Pxr):
        bq = Pxr[:,-1].copy()
        aq = np.abs(Pxr[:,self.n])
        lower_bound = self.bounds[0]*aq
        where_exceeded = bq < lower_bound
        bq[where_exceeded] = lower_bound[where_exceeded]
        upper_bound = self.bounds[1]*aq
        where_exceeded = bq > upper_bound
        bq[where_exceeded] = upper_bound[where_exceeded]
        return bq



"""
    FlowChangeUtils.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import curve_fit
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from molass_legacy.KekLib.ScipyUtils import get_spline
from .Sigmoid import sigmoid, bent_sigmoid, fit_bent_sigmoid, ex_sigmoid
from .FlowChangeCandidates import OUTSTANDING_RATIO_LIMIT

COV_LIMIT = 1e2
L_RATIO_IGMORE_LIMIT = 0.14     # > 0.137 for 20160227, < 0.208 for fc1 in Sugiyama, > 0.13 for 20181212, < 0.144 for 20180316
FIT_HALF_WIDTH = 100
OK_FIT_BALANCE  = 0.2           # > 0.099 for 20170304
OK_K_LIMIT  = 0.1               # < 0.141 for 20170304
RETRY_K_LIMIT = 0.09            # < 0.0975 for 
SAFE_L_RATIO = 0.2              # < 0.27 for Sugiyama
SAFE_FIT_BALANCE = 0.7          # < 0.893 for OA, > 0.673 for , , < 0.95 for OA_Ald
SAFE_DIST_RATIO = 1.0           # > 0.211 for 20190309_1
SAFE_FWD_L_RATIO = 0.25         # > 0.021 for 20180526_OA
SAFE_K_LIMIT = 0.25             # < 0.141 for 20170304, < 0.261 for 20190305_1, < 3.38 for 20161216, < 0.861 for Kosugi3a, < 0.315 for 20170304
SAFE_ERROR_RATIO = 0.7          # > 0.269 for 20161216, > 0.398 for Sugiyama, < 3.83 for 20181127
MUST_ERROR_RATIO = 2            # 
MUST_FIT_BALANCE = 10           # < 13.9 for 20181127, > 9.4 for Sugiyama
FWD_LENGTH_LIMIT = 1.0
OK_TOTAL_ERROR_RATIO = 2.0      # < 2.24 for protains5

def is_ignorable_L_ratio(L_ratio):
    return L_ratio < L_RATIO_IGMORE_LIMIT

def fit_sigmoid_impl(x_, y_, height, peak_region, std_p, ppk, outstanding_ratio, logger, debug=False):
    assert len(x_) > 0

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        print("------- ppk=", ppk)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("fit_sigmoid_impl entry")
            ax.plot(x_, y_)
            for i, p in enumerate(ppk):
                ax.plot(p, y_[p - x_[0]], "o", label=str(i))
            ax.legend()
            plt.show()

    ret_popt = None
    ret_error_ratio = None
    for i, p in enumerate(ppk):
        if debug:
            print([i, p], "---------------------- try")
        try:
            popt, error_ratio = fit_sigmoid_impl_impl(x_, y_, height, peak_region, p, std_p, outstanding_ratio, logger, debug=debug)
        except Exception as exc:
            # to make it robust
            if debug:
                print([i, p], "---------------------- except", exc)
                log_exception(logger, "fit_sigmoid_impl_impl failure: ")
            continue

        L = popt[0]
        L_ratio = abs(L)/height
        if debug:
            print([i, p], "---------------------- error_ratio=%.3g, L_ratio=%.3g" % (error_ratio, L_ratio))
        if ret_error_ratio is None or error_ratio < ret_error_ratio:
            ret_popt = popt
            ret_error_ratio = error_ratio

        if not is_ignorable_L_ratio(L_ratio):
            break

    if i > 0:
        logger.info("fit_sigmoid_impl tried %d times.", i+1)

    return ret_popt, ret_error_ratio

def fit_sigmoid_impl_impl(x_, y_, height, peak_region, x0, std_p, outstanding_ratio, logger, narrower=False, debug=False):
    j0 = x0 - x_[0]
    half_width = FIT_HALF_WIDTH
    if narrower:
        half_width //= 2
    j_start = max(0, j0 - half_width)
    j_end = min(len(y_) - 1, j0 + half_width)
    try:
        y_lower = y_[j_start]
        y_upper = y_[j_end]
    except:
        print((j_start, j_end))
        assert False
    L0 = y_upper - y_lower
    p0 = L0, x0, 1, y_lower

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        with plt.Dp():
            print("narrower=", narrower)
            fig, ax = plt.subplots()
            ax.set_title("fit_sigmoid_impl debug (1)")
            ax.plot(x_, y_)
            ax.plot(x_, sigmoid(x_, *p0))
            fig.tight_layout()
            plt.show()

    fit_slice = slice(j_start, j_end+1)
    fx = x_[fit_slice]
    fy = y_[fit_slice]
    sigmoid_ok = False
    try:
        popt, pcov = curve_fit(sigmoid, fx, fy, p0)
        pcovd = np.diag(pcov)
        if np.isfinite(pcovd[0]):
            error_ratio = abs(np.std(sigmoid(x_, *popt) - y_)/popt[0])
        else:
            # when warned "Covariance of the parameters could not be estimated"
            error_ratio = np.inf

        if False:
            print("popt=", popt)
            print("pcov=", pcovd)
            print("error_ratio=", error_ratio)
            print("std_p=", std_p)
            # assert pcovd[0] < COV_LIMIT
        error_ratio_str = "error_ratio=%.3g" % error_ratio
        sigmoid_ok = True
    except Exception as exc:
        log_exception(logger, "curve_fit(sigmoid, fx, fy, p0) failure: ")
        error_ratio = None
        error_ratio_str = str(exc)

    bent_sigmoid_ok = False
    if (error_ratio is None
        or error_ratio > 0.2                    # error_ratio=0.228 for 20200304_1
        or error_ratio > 0.05 and std_p < 5     # for (*) below
        # (*) as in 20161216 where error_ratio=0.86, or SUB_TRN1 where error_ratio=0.059
        ):
        logger.info('trying bent_sigmoid due to "%s"', error_ratio_str)
        try:
            ret_popt, pcov = fit_bent_sigmoid(fx, fy, x0)
            if np.isfinite(pcov[0,0]):
                error_ratio = abs(np.std(bent_sigmoid(fx, *ret_popt) - fy)/ret_popt[0])
            else:
                error_ratio = np.inf
            s1, s2 = ret_popt[4:]
            print("fit_sigmoid_impl_impl: (s1, s2)=(%.3g, %.3g)" % (s1, s2))
            bent_sigmoid_ok = True
        except:
            log_exception(logger, "fit_bent_sigmoid failure: ")

    if bent_sigmoid_ok:
        pass
    else:
        if sigmoid_ok:
            ret_popt = np.zeros(6)
            ret_popt[0:4] = popt
        else:
            raise Exception("No fits at x0=%.4g" % (x0))

    L, x0, k = ret_popt[0:3]
    dist_ratio = peak_region.get_distance_ratio(x0)
    fwd_slice = slice(int(round(x0 - fx[0])), None)
    ffx = fx[fwd_slice]
    ffy = fy[fwd_slice]
    try:
        spline = get_spline(ffx, ffy)
        sffy = spline(ffx)
        fity = ex_sigmoid(ffx, *ret_popt)
        fwd_L_ratio = np.std(sffy - fity)/abs(L)
    except:
        log_exception(logger, "get_spline: ")
        # as in Sugiyama
        sffy = ffy          # for debug
        fwd_L_ratio = 0     # to avoid to be discarded

    fit_balance = abs((x0 - fx[0])/(fx[-1] - x0) - 1)

    total_error_ratio = abs(np.std(ex_sigmoid(x_, *ret_popt) - y_)/ret_popt[0])

    if debug:
        print("narrower=", narrower)
        print("error_ratio=%.3g, fwd_L_ratio=%.3g" % (error_ratio, fwd_L_ratio))
        print("total_error_ratio=%.3g" % total_error_ratio)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("fit_sigmoid_impl debug (2)")
            ax.plot(x_, y_)
            ax.plot(x_, ex_sigmoid(x_, *ret_popt))
            ax.plot(ffx, sffy)
            ax.plot(ffx, ex_sigmoid(ffx, *ret_popt))
            tx = np.average(ax.get_xlim())
            ty = np.average(ax.get_ylim())
            ax.text(tx, ty, "fit_balance=%.3g, k=%.3g, error_ratio=%.3g, fwd_L_ratio=%.3g" % (fit_balance, k, error_ratio, fwd_L_ratio), ha="center", alpha=0.5, fontsize=16)
            fig.tight_layout()
            plt.show()

    reasons = ""
    def add_reason(reason):
        nonlocal reasons
        if reasons == "":
            reasons += "reasons"
        reasons += reason

    if total_error_ratio > OK_TOTAL_ERROR_RATIO:
        add_reason("(-1) total_error_ratio=%.3g > %.3g" % (total_error_ratio, OK_TOTAL_ERROR_RATIO))

    L_ratio = abs(L)/height

    if error_ratio > MUST_ERROR_RATIO or fit_balance > MUST_FIT_BALANCE:
        add_reason(" (0) error_ratio=%.3g > %.3g or fit_balance=%.3g > %.3g" % (error_ratio, MUST_ERROR_RATIO, fit_balance, MUST_FIT_BALANCE))

    if k > SAFE_K_LIMIT:
        if error_ratio > SAFE_ERROR_RATIO:
            add_reason(" (1) error_ratio=%.3g > %.3g" % (error_ratio, SAFE_ERROR_RATIO))
        if L_ratio > SAFE_L_RATIO:
            # as in Sugiyama
            pass
        else:
            if fit_balance > SAFE_FIT_BALANCE:
                # as in OA_Ald right
                print("L_ratio=", L_ratio)
                add_reason(" (2) fit_balance=%.3g > %.3g" % (fit_balance, SAFE_FIT_BALANCE))
    else:
        if (k > OK_K_LIMIT
            and fit_balance < OK_FIT_BALANCE
            and outstanding_ratio > OUTSTANDING_RATIO_LIMIT     # required to exclude 20200211
            ):
            # as in retry in 20170304
            pass
        else:
            if not narrower and k > RETRY_K_LIMIT:
                # as in 20190305_1
                try:
                    p = int(round(x0))
                    ret_popt, error_ratio = fit_sigmoid_impl_impl(x_, y_, height, peak_region, p, std_p, outstanding_ratio, logger, narrower=True, debug=False)
                    logger.info("retry successful with k=%.3g", k)
                except Exception as exc:
                    raise exc
            else:
                add_reason(" (3) k=%.3g < %.3g" % (k, SAFE_K_LIMIT))
                if fit_balance > SAFE_FIT_BALANCE:
                    add_reason(" (4) fit_balance=%.3g > %.3g" % (fit_balance, SAFE_FIT_BALANCE))
                elif dist_ratio < SAFE_DIST_RATIO and fwd_L_ratio > SAFE_FWD_L_RATIO:
                    add_reason(" (5) dist_ratio=%.3g < =%.3g " % (dist_ratio, SAFE_DIST_RATIO))

    if fwd_L_ratio > SAFE_FWD_L_RATIO:
        add_reason(" (6) fwd_L_ratio=%.3g > %.3g" % (fwd_L_ratio, SAFE_FWD_L_RATIO))

    if reasons != "":
        logger.info("candidate at %.4g will be discarded due to %s", x0, reasons)
        raise Exception("Bad fit at x0=%.4g: %s" % (x0, reasons))

    return ret_popt, error_ratio



"""
    Stochastic.NdSphericalVolume.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from matplotlib.patches import Rectangle, Circle
from scipy.special import gamma
import seaborn as sns
sns.set_theme()

fig = plt.figure(figsize=(18,4))
ax0 = fig.add_subplot(141)
ax1 = fig.add_subplot(142, projection='3d')
ax2 = fig.add_subplot(143)
ax3 = fig.add_subplot(144)
ax3.set_yscale('log')

ax0.set_title(r"Area Ratio of 2D Circle/Square: $\pi\cdot 2^{-2} \approx %.2g$" % (np.pi/4))
ax0.set_aspect('equal')
square = Rectangle((-1,-1), 2, 2, color='red', alpha=0.1)
ax0.add_patch(square)
circle = Circle(xy=(0,0), radius=1.0, color='blue', alpha=0.1)
ax0.add_patch(circle)
ax0.set_xlim(-1.2, 1.2)
ax0.set_ylim(-1.2, 1.2)

v3 = 4/3*np.pi
ax1.set_title(r"Volume Ratio of 3D Sphere/Cube: $\frac{4}{3}\pi \cdot 2^{-3} \approx %.2g$" % (v3 * 2**(-3)))

# See https://stackoverflow.com/questions/73679251/how-can-i-plot-a-cube-in-matplotlibpython-when-the-8-vertices-are-given-in-num
# Unit cube coordinates
x = np.array([0, 1, 1, 0, 0, 1, 1, 0])*2 - 1
y = np.array([0, 0, 1, 1, 0, 0, 1, 1])*2 - 1
z = np.array([0, 0, 0, 0, 1, 1, 1, 1])*2 - 1

# Face IDs
vertices = [[0,1,2,3],[1,5,6,2],[3,2,6,7],[4,0,3,7],[5,4,7,6],[4,5,1,0]]

tupleList = list(zip(x, y, z))

poly3d = [[tupleList[vertices[ix][iy]] for iy in range(len(vertices[0]))] for ix in range(len(vertices))]
# ax1.scatter(x,y,z)
ax1.add_collection3d(Poly3DCollection(poly3d, facecolors='r', linewidths=1, alpha=0.1))

u = np.linspace(0, 2 * np.pi, 100)
v = np.linspace(0, np.pi, 100)
x = np.outer(np.cos(u), np.sin(v))
y = np.outer(np.sin(u), np.sin(v))
z = np.outer(np.ones(np.size(u)), np.cos(v))

# Plot the surface
ax1.plot_surface(x, y, z)

# Set an equal aspect ratio
ax1.set_aspect('equal')

x_ = np.arange(2,40)
y_ = np.power(np.pi, x_/2)/gamma(x_/2 + 1) * np.power(2.0, -x_)

for ax, scale in [(ax2, 'Linear'), (ax3, 'Log')]:
    ax.set_title(r"Volume Ratio of N-D Sphere/Cube (%s scale)" % scale)
    ax.set_xlabel("Dimension: N")
    ax.set_ylabel("Volume")
    ax.plot(x_, y_, label=r"$VR=\frac{\pi^{n/2}}{\Gamma(n/2 + 1)} \cdot 2^{-n}$")
    ax.plot(3, v3/8, "o", color="red", label="3D Sphere/Cube")
    ax.legend()

fig.tight_layout()
plt.show()


"""

    StageExtrapolation.py

    Copyright (c) 2016-2023, SAXS Team, KEK-PF

"""
import os
import numpy as np
from scipy import stats
from bisect import bisect_right
from openpyxl import Workbook
from molass_legacy.KekLib.NumpyUtils import np_savetxt, get_valid_index, simply_safe_sprintf
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry, ordinal_str 
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker, log_exception
from molass_legacy.AutorgKek.KekToolsGP import AutorgKek, ErrorResult
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.AutorgKek.AtsasTools import autorg as autorg_atsas
from molass_legacy.ATSAS.Almerge import AlmergeExecutor
from molass_legacy.ATSAS.DatGnom import DatgnomExecutor
from ZeroExtrapolator import ZeroExtrapolator, GuinierPorodAnalyzer, NUM_EXTRAPOLATION_POINTS
from molass_legacy.Reports.ZeroExtrapolationResultBook import divide
from molass_legacy.Reports.DatgnomResultBook import make_datgnom_result_book
from molass_legacy.KekLib.ProgressInfo import put_info, put_max_info, put_error
from molass_legacy.SerialAnalyzer.ProgressInfoUtil import extimate_zero_extrapolation, STREAM_ZERO_EX, NUM_SHEET_TYPES
from molass_legacy.SerialAnalyzer.LinearityScore import LinearityScore
from molass_legacy.SerialAnalyzer.DevSettings import get_dev_setting
from molass_legacy.Reports.ReportUtils import make_zero_extrapolation_book, make_zero_extrapolation_overlay_book
from molass_legacy.Test.TesterLogger import write_to_tester_log
from molass_legacy.GuinierAnalyzer.AutorgKekAdapter import AutorgKekAdapter
from molass_legacy.Reports.DefaultFont import set_default_font
set_default_font()

USE_SIMPLIFIED_GUINIER_POROD    = False
TEST_LOG_AQ_GAP     = False
LARGE_FLOAT_VALUE   = 0     # this is temporary. fix this case

def control_extrapolation( self ):
    if len(self.applied_ranges) > 0:
        prepare_extrapolation( self )
        # log_peak_info( self )
        try:
            do_extrapolation(self)
            clean_tempfolders(self)
        except Exception as exc:
            etb = ExceptionTracebacker()
            self.logger.error( etb )
            put_error( self.stream )
    else:
        self.logger.warning( 'No range for zero concentration extrapolation was found.' )

def log_peak_info( self ):
    for m, range_ in enumerate( self.applied_ranges ):
        write_to_tester_log( 'range_info=' + str( ( m,  range_) ) + '\n' )
        middle = range_.get_fromto_list()[0][1]
        """
            this middle is not appropriate for the new style PairedRange's
            when the ascending side and the descending side are made apart.
        """
        result = self.guinier_result_array[middle][1]
        # TODO: result.fit.Rg is not appropriate when restored
        write_to_tester_log( 'peak_rg_info=' + str( ( m, result.Rg, result.fit.Rg ) ) + '\n' )

def prepare_extrapolation( self ):

    self.to_extrapolate_intensity_array = []
    self.rg_vector_array = []
    self.iz_vector_array = []
    self.indeces_array = []

    to_extrapolate = {}

    for m, range_ in enumerate( self.applied_ranges ):
        self.stop_check()

        from_to_list = range_.get_fromto_list()

        indeces_pair = []

        for j, fromto in enumerate(from_to_list):
            f, t = fromto
            indeces = np.array( np.arange( f, t+1 ), dtype=int )
            print( 'indeces[%d]=' % m, indeces )
            indeces_pair.append( indeces )
            for i in indeces:
                key_list = to_extrapolate.get(i)
                if key_list is None:
                    to_extrapolate[i] = key_list = []
                key_list.append( (m, j) )
                # key_list contais two tuples only for i=middle in old style

        self.indeces_array.append(indeces_pair)
        self.to_extrapolate_intensity_array.append( [ [], [] ] )
        self.rg_vector_array.append( [ [], [] ] )
        self.iz_vector_array.append( [ [], [] ] )

    # self.logger.info('prepare_extrapolation: to_extrapolate=%s', str(to_extrapolate))
    """
    for MCT-20190522,
    to_extrapolate=
        {0: [(0, 0), (1, 0)],
         1: [(0, 0), (1, 0)],
         2: [(0, 0), (1, 0)],
            ...
        54: [(0, 0), (1, 0)]}
    """

    self.guinier_boundaries_dict = {}
    for i, rec in enumerate( self.guinier_result_array ):
        intensity   = rec[0]
        result_kek  = rec[1]
        result_atsas = rec[2]
        assert result_atsas is not None

        key_list = to_extrapolate.get(i)
        if key_list is None:
            continue

        # key_list contais two tuples only for the peak, and one tuple for others.
        for m, j in key_list:
            self.to_extrapolate_intensity_array[m][j].append( intensity )
            Rg_stdev = 0 if result_kek.Rg_stdev is None else result_kek.Rg_stdev
            if np.isinf( Rg_stdev ):
                Rg_stdev = LARGE_FLOAT_VALUE
            self.rg_vector_array[m][j].append( [ result_kek.Rg, Rg_stdev, result_atsas.Rg, result_atsas.Rg_stdev ] )
            I0_stdev = 0 if result_kek.I0_stdev is None else result_kek.I0_stdev
            if np.isinf( I0_stdev ):
                I0_stdev = LARGE_FLOAT_VALUE
            self.iz_vector_array[m][j].append( [ result_kek.I0, I0_stdev, result_atsas.I0, result_atsas.I0_stdev ] )

            guinier_boundary = self.guinier_boundaries_dict.get( (m, j) )
            if guinier_boundary is None:
                self.guinier_boundaries_dict[(m,j)] = [ result_kek.To ]
            else:
                if result_kek.To is None:
                    pass
                else:
                    if guinier_boundary[0] is None or result_kek.To > guinier_boundary[0]:
                        guinier_boundary[0] = result_kek.To

    # self.logger.info('prepare_extrapolation: guinier_boundaries_dict=%s', str(self.guinier_boundaries_dict))

    max_dict = extimate_zero_extrapolation( self.applied_ranges )
    # self.logger.info( 'max_dict=' + str(max_dict) )
    put_max_info( STREAM_ZERO_EX, max_dict )

def do_extrapolation(self, debug=False):
    if debug:
        self.logger.info("Performing do_extrapolation...")
    qvector = self.qvector      # self.qvector has been set at the start of run_gunier_analysis
    bq_sum_start = bisect_right( qvector, 0.2 )

    zx = ZeroExtrapolator( qvector, self.preview_params, self.serial_data, self.mapped_info, self.applied_ranges, self.known_info_list)

    if USE_SIMPLIFIED_GUINIER_POROD:
        q_slice = slice( 30, len(qvector)//4 )
        x_ = qvector[q_slice]

    if self.using_avareged_files:
        self.used_datafiles = self.averaged_datafiles
    else:
        self.used_datafiles = self.serial_data.datafiles

    tester_zx_save  = get_dev_setting( 'tester_zx_save' )
    if tester_zx_save == 1 or self.maintenance_mode:
        self.zx_out_folder  = self.work_folder + '/extrapolated'
        zx_save = True
    else:
        self.zx_out_folder  = None
        zx_save = False

    if zx_save and not os.path.exists( self.zx_out_folder ):
        mkdirs_with_retry( self.zx_out_folder )

    if self.atsas_is_available:
        self.almerge = AlmergeExecutor()
        self.datgnom = DatgnomExecutor()
        if self.excel_is_available:
            self.result_wb_datgnom = None
        else:
            self.result_wb_datgnom = Workbook()
            self.result_wb_datgnom.remove_sheet(self.result_wb_datgnom.active)
    else:
        self.almerge = None
        self.datgnom = None

    self.logger.info( 'extrapolating for ' + str(self.report_ranges) )
    self.peak_num_ranges_list = []

    ignore_all_bqs = get_setting('ignore_all_bqs')
    ignore_bq_list = get_setting('ignore_bq_list')

    self.guinier_results = []
    range_no = 0
    for m, range_ in enumerate( self.applied_ranges ):
        data_array = []
        boundary_indeces = []
        param_array_list = []
        fromto_list = range_.get_fromto_list()
        peak_num_ranges = len(fromto_list)
        self.peak_num_ranges_list.append(peak_num_ranges)
        ignore_bq_for_overlay_book = []
        for j, fromto in enumerate(fromto_list):
            intensities = self.to_extrapolate_intensity_array[m][j]
            log_message = 'extrapolation start for the %s peak ad(%d) with range %s' % ( ordinal_str(m), j, str(fromto) )
            self.logger.info( log_message )
            write_to_tester_log( log_message + '\n' )
            self.stop_check()

            if self.doing_sec:
                if peak_num_ranges == 1 and self.range_type != 5:
                    ad = 'both-side'
                else:
                    ad = 'asc-side' if j == 0 else 'desc-side'
            else:
                ad = 'nat-comp' if m == 0 else 'unf-comp'

            if zx_save:
                paren_no = '' if len(self.applied_ranges ) == 1 else '(%d)' % (m + 1)
                zx_out_sub_folder = self.zx_out_folder + '/' + ad + paren_no
                if not os.path.exists( zx_out_sub_folder ):
                    mkdirs_with_retry( zx_out_sub_folder )

            try:
                indeces = self.indeces_array[m][j]
            except:
                # TOODO: better control
                continue

            c_vector = self.c_vector[indeces]
            # print( 'c_vector=', c_vector )

            try:
                guinier_boundary = self.guinier_boundaries_dict.get( (m, j) )[0]
            except Exception as exc:
                # as in 20161119/Kosugi3a_Backsub
                self.logger.warning( 'failed to get guinier_boundary for (m, j)=' + str( (m, j) ) )
                for k, v in sorted( self.guinier_boundaries_dict.items() ):
                    self.logger.info( str( (k, v) ) ) 
                raise exc

            self.logger.info( 'Guinier boundary is assumed at Q[%d]=%.3g.' % ( guinier_boundary, qvector[guinier_boundary] ) )

            try:
                cx_vector, min_c, max_c = zx.make_cx_vector( j, range_, indeces, c_vector )
                zx_result, lrf_info = zx.extrapolate( m, j, range_no, intensities, c_vector,
                                                        cx_vector, max_c,
                                                        guinier_boundary=guinier_boundary,
                                                        temp_folder=self.temp_folder )

                data_matrix_c, ze_array, param_array, ab_error_array, ze_error_array, boundary_j = zx_result
            except Exception as exc:
                etb = ExceptionTracebacker()
                self.logger.error( str(etb) )
                raise exc

            data_array.append( [ data_matrix_c, ze_array ] )
            boundary_indeces.append( boundary_j )
            param_array_list.append( param_array )

            if debug:
                self.logger.info("put_info( ( %d, %d ), %d )", STREAM_ZERO_EX, m*NUM_SHEET_TYPES, 1)
            put_info( ( STREAM_ZERO_EX, m*NUM_SHEET_TYPES+j ), 1 )
            self.logger.info( ordinal_str(m+1) + ' peak ' + ad + ' extrapolation done.' )

            izx_vector = []
            rgx_vector = []
            rg_results = []
            for k in range( ze_array.shape[1] ):
                self.stop_check()
                c = cx_vector[k]

                if USE_SIMPLIFIED_GUINIER_POROD:
                    y_ = ze_array[q_slice, k]
                    gp = GuinierPorodAnalyzer( x_, y_ )
                    G, G_stdev, Rg, Rg_stdev = gp.fit()
                else:
                    ze_intensity = np.vstack( [ qvector, ze_array[:, k], ze_error_array[:, k] ] ).T
                    try:
                        if self.use_simpleguinier == 0:
                            autorg_kek = AutorgKek( ze_intensity )
                        else:
                            autorg_kek = AutorgKekAdapter( ze_intensity )
                        result = autorg_kek.run()
                        G, G_stdev, Rg, Rg_stdev = result.I0, result.I0_stdev, result.Rg, result.Rg_stdev
                    except Exception as exc:
                        etb = ExceptionTracebacker()
                        self.logger.error( etb )
                        G, G_stdev, Rg, Rg_stdev = None, 0, None, 0
                        result = None

                    if zx_save:
                        zx_filename = make_zx_filename( self, j, indeces, k, c )
                        np_savetxt( zx_out_sub_folder + '/' +zx_filename, ze_intensity )

                rg_results.append(result)

                if G_stdev is None:
                    G_stdev = 0
                if np.isinf( G_stdev ):
                    G_stdev = LARGE_FLOAT_VALUE

                if Rg_stdev is None:
                    Rg_stdev = 0
                if np.isinf( Rg_stdev ):
                    Rg_stdev = LARGE_FLOAT_VALUE

                izx_vector.append( [ G, G_stdev ] )
                rgx_vector.append( [ Rg, Rg_stdev ] )
                if debug:
                    self.logger.info("put_info( ( %d, %d ), %d )", STREAM_ZERO_EX, m*NUM_SHEET_TYPES+j, 2 + k)
                put_info( ( STREAM_ZERO_EX, m*NUM_SHEET_TYPES+j ), 2 + k )

            if self.log_memory_usage == 1:
                m_info = self.process.memory_info()
                memory_usage    = " " + str( ( m_info.vms, ) )
            else:
                memory_usage    = ""

            # self.logger.info( str([k]) + 'rgx_vector=' + str(rgx_vector) )

            zx_index = 0 if j == 0 else -1    # at zero-conc
            self.guinier_results.append(rg_results[zx_index])
            zx_rg_z, zx_rg_z_error = rgx_vector[ zx_index ]
            zx_iz_z, zx_iz_z_error = izx_vector[ zx_index ]
            zx_index = -1 if j == 0 else 0    # at peak
            zx_rg_p = rgx_vector[ zx_index ][0]
            zx_conc_p = cx_vector[ zx_index ]

            rg_vector = self.rg_vector_array[m][j]
            iz_vector = self.iz_vector_array[m][j]
            c_rg_iz = [ [ c_vector,  cx_vector ],
                        [ np.array(rg_vector), np.array(rgx_vector) ],
                        [ np.array(iz_vector), np.array(izx_vector) ],
                      ]

            if self.atsas_is_available:
                make_atsas_folder( self )
                self.atsas_exz_file = make_filename( self, m, j, peak_num_ranges, '/atsas', '.dat' )
                try:
                    # self.logger.info( '%d-%d indeces=' % (m,j) + str(indeces) )
                    almerge_result = self.almerge.execute( self.c_vector, self.used_datafiles, indeces, self.atsas_exz_file )
                except:
                    etb = ExceptionTracebacker()
                    etb.log()
                    almerge_result = None
                if almerge_result is None:
                    result_atsas, result_atsas_eval = ErrorResult(), ErrorResult()
                    overlap_from_max = None
                else:
                    overlap_from_max_ = almerge_result.overlap_from_max
                    if overlap_from_max_ >= len(qvector):
                        overlap_from_max_ = len(qvector) - 1
                        write_to_tester_log( 'overlap_from_max_ was changed to %d\n' % ( overlap_from_max_ ) )
                    overlap_from_max = qvector[overlap_from_max_]

                    result_atsas, result_atsas_eval = autorg_atsas( self.atsas_exz_file )

                if result_atsas is not None:
                    self.logger.info( ordinal_str(m+1) + ' peak ' + ad
                                        + ' atsas result: Rg=' + str(result_atsas.Rg)
                                        + ' I(0)=' + str(result_atsas.I0) )

                write_to_tester_log( simply_safe_sprintf( 'compare atsas Rg: m=%d j=%d Rg=%.17g Rg=%.17g Rg_stdev=%.17g Rg_stdev=%.17g\n',
                                                            m, j, zx_rg_z, result_atsas.Rg, zx_rg_z_error, result_atsas.Rg_stdev ) )
                write_to_tester_log( simply_safe_sprintf( 'compare atsas Iz/c: m=%d j=%d Iz/c=%.17g Iz/c=%.17g Iz/c_stdev=%.17g Iz/c_stdev=%.17g\n',
                                                            m, j, zx_iz_z, divide( result_atsas.I0, zx_conc_p, None), zx_iz_z_error, divide( result_atsas.I0_stdev, zx_conc_p, 0 ) ) )
            else:
                almerge_result  = None
                result_atsas    = None
                overlap_from_max= None

            try:
                write_to_tester_log( simply_safe_sprintf( 'compare zero-peak: m=%d j=%d Rg=%g Rg=%g\n', m, j, zx_rg_z, zx_rg_p ) )
                bq_max_q = np.nan if boundary_j is None else qvector[boundary_j]
                write_to_tester_log( simply_safe_sprintf( 'bq_max=%g bq_boundary=%g\n', zx.bq_max, bq_max_q ) )

                zx_rg_str = simply_safe_sprintf( '%.3g ', zx_rg_z )
                self.logger.info( ordinal_str(m+1) + ' peak ' + ad + ' rg re-estimation done with extrapolated Rg=' + zx_rg_str + memory_usage )

                # Rg-extrapolation linearity
                lin_score = LinearityScore( cx_vector, np.array(rgx_vector)[:,0] )
                rg_slope, rg_r_value, rg_std_err = lin_score.get_params()

                # Iz-extrapolation linearity
                lin_score = LinearityScore( cx_vector, np.array(izx_vector)[:,0] )
                iz_slope, iz_r_value, iz_std_err = lin_score.get_params()

                if zx_rg_z is None or zx_iz_z is None:
                    write_to_tester_log( 'WARNING: zx_rg_z=' + str(zx_rg_z) + ', zx_iz_z=' + str(zx_iz_z) + '\n' )
                else:
                    write_to_tester_log( simply_safe_sprintf( 'extrapolation linearity: m=%d j=%d  rg_slope=%g rg_r_value=%g rg_std_err=%g iz_slope=%g iz_r_value=%g iz_std_err=%g min_c=%g max_c=%g overlap_from_max=%s\n',
                                                                m, j, rg_slope, rg_r_value, rg_std_err/zx_rg_z, iz_slope, iz_r_value, iz_std_err/zx_iz_z, min_c, max_c, str(overlap_from_max) ) )

                if TEST_LOG_AQ_GAP:
                    if boundary_j is None:
                        aq_gap = np.nan
                    else:
                        xb = qvector[boundary_j]
                        yb = []
                        for slice_ in [ slice(boundary_j-60, boundary_j), slice(boundary_j,boundary_j+60) ]:
                            x = qvector[slice_]
                            y = param_array[slice_, 0]
                            slope, intercept, _, _, _ = stats.linregress( x, y )
                            yb.append( slope*xb + intercept )
                        aq_gap = yb[1] - yb[0]

                    bq_sum = np.sum( param_array[bq_sum_start:, 1] )
                    write_to_tester_log( 'bq_sum=%g aq_gap=%g\n' % ( bq_sum, aq_gap ) )

                # TODO: pH7 error

                fromto_list = range_.get_fromto_list()
                from_, to_ = fromto_list[j]
                self.zx_summary_list.append( [ m, j, from_, to_, to_ - from_ + 1, min_c, max_c,
                                                zx_rg_z, zx_rg_z_error, zx_iz_z, zx_iz_z_error,
                                                rg_r_value, rg_std_err, iz_r_value, iz_std_err,
                                                ] )

            except Exception as exc:
                etb = ExceptionTracebacker()
                self.logger.error( etb )

            # output_smoothing has been deprecated

            param_array_out = param_array
            orig_param_array = None

            num_zx_points   = ze_array.shape[1]
            mjn = ( m, j, len(self.applied_ranges), peak_num_ranges )

            if ignore_bq_list is None:
                ignore_bq = ignore_all_bqs
            else:
                ignore_bq = ignore_bq_list[range_no]
            ignore_bq_for_overlay_book.append(ignore_bq)

            try:
                if self.excel_is_available:
                    wb = Workbook()
                    ws = wb.active
                else:
                    wb = self.result_wb
                    ws = wb.create_sheet('Extrapolation')

                book_file = self.temp_folder + '/--zero_extrapolation-temp-%d-%d.xlsx' % ( m, j )
                make_zero_extrapolation_book( self.excel_is_available, wb, ws, book_file, mjn,
                                                indeces, c_rg_iz, self.serial_data.qvector,
                                                data_matrix_c, self.xr_j0,
                                                ze_array, param_array_out, orig_param_array,
                                                guinier_boundary,
                                                almerge_result, result_atsas,
                                                boundary_j, parent=self, lrf_info=lrf_info,
                                                logger=self.logger )
                if debug:
                    self.logger.info("put_info( ( %d, %d ), %d )", STREAM_ZERO_EX, m*NUM_SHEET_TYPES+j, 2 + num_zx_points)
                put_info( ( STREAM_ZERO_EX, m*NUM_SHEET_TYPES+j ), 2 + num_zx_points )
                self.temp_books.append( book_file )

                if self.log_memory_usage == 1:
                    m_info = self.process.memory_info()
                    memory_usage    = " " + str( ( m_info.vms, ) )
                else:
                    memory_usage    = ""
                self.logger.info( ordinal_str(m+1) + ' peak ' + ad + ' book done.' + memory_usage )
            except Exception as exc:
                etb = ExceptionTracebacker()
                self.logger.error( etb )

            output_extrapolation_params( self, m, j, peak_num_ranges, qvector, param_array_out, ab_error_array, indeces, ignore_bq)
            range_no += 1

            if self.atsas_is_available:
                try:
                    datgnom_out_file = make_filename( self, m, j, peak_num_ranges, '/atsas', '.out' )
                    datgnom_result = self.datgnom.execute( self.zx_a_file_path, zx_rg_z, datgnom_out_file )
                    self.zx_summary_list2.append( [ datgnom_result.RgPr, datgnom_result.RgPr_Error, datgnom_result.IzPr, datgnom_result.IzPr_Error, datgnom_result.Dmax, *datgnom_result.Qrange ] )
                    write_to_tester_log( 'datgnom_result=' + str( self.zx_summary_list2[-1] ) + '\n' )
                    book_file = self.temp_folder + '/--zero_extrapolation-atsas-%d-%d.xlsx' % ( m, j )
                    try:
                        if self.excel_is_available:
                            wb = Workbook()
                            ws = wb.active
                        else:
                            wb = self.result_wb_datgnom
                            ws = wb.create_sheet('DATGNOM')

                        make_datgnom_result_book( self.excel_is_available, wb, ws,
                                        mjn, self.range_type, datgnom_out_file, book_file, self)
                        self.temp_books_atsas.append( book_file )
                        self.logger.info( ordinal_str(m+1) + ' peak ' + ad + ' atsas book done.' )
                    except Exception as exc:
                        self.logger.error( 'make_datgnom_result_book failed with ' + str(exc) )
                except Exception as exc:
                    log_exception(self.logger, "datgnom failure: ")
                    self.zx_summary_list2.append( [None] * 5 )
                if debug:
                    self.logger.info("put_info( ( %d, %d ), %d )", STREAM_ZERO_EX, m*NUM_SHEET_TYPES+j, 3 + num_zx_points)
                put_info( ( STREAM_ZERO_EX, m*NUM_SHEET_TYPES+j ), 3 + num_zx_points )
            else:
                self.zx_summary_list2.append( [None] * 5 )

        if peak_num_ranges == 2:
            ignore_bq = np.sum(ignore_bq_for_overlay_book) == 2
            try:
                if self.excel_is_available:
                    wb = Workbook()
                    ws = wb.active
                else:
                    wb = self.result_wb
                    ws = wb.create_sheet('Extrapolation')

                book_file = self.temp_folder + '/--zero_extrapolation-temp-%d-%d.xlsx' % ( m, 2 )
                self.temp_books.append( book_file )
                make_zero_extrapolation_overlay_book( self.excel_is_available, wb, ws,
                                                        book_file, ( m, len(self.applied_ranges) ),
                                                        self.serial_data.qvector, param_array_list, boundary_indeces,
                                                        parent=self,
                                                        lrf_info=lrf_info,
                                                        logger=self.logger )
                if debug:
                    self.logger.info("put_info( ( %d, %d ), %d )", STREAM_ZERO_EX, m*NUM_SHEET_TYPES+2, 1)
                put_info( ( STREAM_ZERO_EX, m*NUM_SHEET_TYPES+2 ), 1 )

                if self.log_memory_usage == 1:
                    m_info = self.process.memory_info()
                    memory_usage    = " " + str( ( m_info.vms, ) )
                else:
                    memory_usage    = ""
                self.logger.info( ordinal_str(m+1) + ' peak overlay book done.' + memory_usage )
            except Exception as exc:
                etb = ExceptionTracebacker()
                self.logger.error( etb )

def make_atsas_folder( self ):
    self.atsas_folder = self.work_folder + '/atsas'
    if not os.path.exists( self.atsas_folder ):
        mkdirs_with_retry( self.atsas_folder )

def make_filename( self, m, ud, peak_num_ranges, sub_folder, tail ):
    if self.doing_sec:
        if peak_num_ranges == 1:
            comp = '_bth'
        else:
            comp = '_asc' if ud == 0 else '_dsc'
    else:
        comp = '_nat' if m == 0 else '_unf'
    filename = self.work_folder + sub_folder + '/pk%d' % (m+1) + comp + tail
    return filename

def make_zx_filename( self, j, indeces, k, c ):
    conc_str = ( '%.4g' % c ).replace( '.', '' )[-4:]
    if j == 0:
        if k < NUM_EXTRAPOLATION_POINTS:
            fno = None
        else:
            fno = indeces[k-NUM_EXTRAPOLATION_POINTS]
    else:
        if k < len( indeces ):
            fno = indeces[k]
        else:
            fno = None
    fno_str = '____' if fno is None else '%05d' % fno
    return 'extrapolated-%03d-%s-%s.dat' % ( k, fno_str, conc_str )

def output_extrapolation_params( self, m, ud, peak_num_ranges, qvector, param_array, ab_error_array, indeces, ignore_bq ):
    self.zx_a_file_path = make_filename( self, m, ud, peak_num_ranges, '', '_A_cn.dat' )
    self.zx_b_file_path = make_filename( self, m, ud, peak_num_ranges, '', '_B_cn.dat',  )

    # qvector_ = np.atleast_2d( qvector ).T
    filepaths = [ self.zx_a_file_path, self.zx_b_file_path ]

    num_files = 1 if ignore_bq else 2
    for j in range(num_files):
        values = param_array[:,j]
        if j > 0:
            if np.sum(values) == 0:     # task: need a more effective way to judge this
                continue

        fh = open( filepaths[j], "wb" )
        fh.write( str.encode( "# Extrapolated using the following files with []-numbering starting from 1.\n" ) )
        for k in indeces:
            fh.write( str.encode( "# [%d] %s\n" % ( k+1, self.used_datafiles[k] )  ) )
        fh.write( str.encode( "#\n# Q\tIntensity\tError\n" ) )
        fh.close()
        array = np.vstack( [ qvector, values, ab_error_array[:,j] ] ).T
        np_savetxt( filepaths[j], array, mode="a" )

def clean_tempfolders(self):
    averaged_data_folder = get_setting('averaged_data_folder')  # this should have set in AnalyzerDialogProxy
    if os.path.exists(averaged_data_folder):
        keep_tempfolder_averaged = get_setting("keep_tempfolder_averaged")
        if not keep_tempfolder_averaged:
            import shutil
            try:
                # assert False
                shutil.rmtree(averaged_data_folder)
                self.logger.info("%s has been removed.", averaged_data_folder)
            except:
                log_exception(self.logger, "averaged_data_folder cleanup failed: ")



# coding: utf-8
"""
    Optimizer.Demo.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
from time import sleep
import numpy as np
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
from molass_legacy._MOLASS.SerialSettings import set_setting
from CorrectedData import CorrectedXray
import molass_legacy.KekLib.DebugPlot as plt
from .OptimalElution import compute_optimal_elution
from MatrixData import simple_plot_3d
from ModeledData import ModeledData
from Tutorials.SvdPrecision import plot_text
from .GlobalInspector import FIG_SIZE, GlobalInspector

def demo(in_folder, sp=None, rank=None, debug=False, log_fh=None, fig_file=None, data_no=None, logger=None):
    print(in_folder)
    set_setting('in_folder', in_folder)

    if sp is None:
        xr = CorrectedXray(in_folder)
        D = xr.data
        E = xr.error
        qvector = xr.vector
        ecurve = xr.ecurve
        legacy_info = None
    else:
        from DecompUtils import CorrectedBaseline
        sp.load(in_folder)
        sd = sp.get_corrected_sd()
        D, E, qvector, ecurve = sd.get_xr_data_separate_ly()
        corbase_info = CorrectedBaseline(sd, sp.mapper)
        legacy_info = [sd, sp.mapper, corbase_info]

    fig = plt.figure(figsize=FIG_SIZE)

    # inspector_impl(fig, in_folder, D, E, qvector, ecurve, log_fh=log_fh, data_no=data_no, legacy_info=legacy_info, logger=logger)
    gi = GlobalInspector(D, E, ecurve, legacy_info, logger)

    if log_fh is not None:
        log_fh.write(','.join([str(data_no), in_folder] + ["%.3g" % r for r in gi.norm_ratios]) + "\n")
        log_fh.flush()

    gi.plot_results(fig, D, qvector)

    if debug:
        plt.show()
    else:
        if fig_file is not None:
            fig.savefig( fig_file )

        plt.show(block=False)
        sleep(0.5)

A_SIZE = 600
E_SIZE = 300
MATRIX_SYM_FONTSIZE = 200

def method_proof():
    q = np.linspace(0.01, 0.6, A_SIZE)
    rg_list = [50, 24]
    d_list = [1, 3]
    md = ModeledData(q, E_SIZE, rg_list=rg_list, d_list=d_list)
    M = md.get_data()

    fig = plt.figure(figsize=(21,12))
    gs = GridSpec(3,3)
    ax00 = fig.add_subplot(gs[0,0], projection='3d')
    ax01 = fig.add_subplot(gs[0,1])
    ax02 = fig.add_subplot(gs[0,2])
    ax10 = fig.add_subplot(gs[1,0], projection='3d')
    ax11 = fig.add_subplot(gs[1,1])
    ax12 = fig.add_subplot(gs[1,2])
    ax22 = fig.add_subplot(gs[2,2])

    fig.suptitle("Proof of the Factored Weights Method", fontsize=30)

    simple_plot_3d(ax00, M, x=q)

    for c in md.C:
        ax01.plot(c)
    plot_text(ax01, "$C$", fontsize=MATRIX_SYM_FONTSIZE)

    for k, p in enumerate(md.P.T):
        ax02.plot(q, p, color='C%d'%k)

    P_ = M @ np.linalg.pinv(md.C)
    for k, p in enumerate(P_.T):
        ax02.plot(q, p, ':', color='C%d'%k)
    plot_text(ax02, "$P$", fontsize=MATRIX_SYM_FONTSIZE)

    aw = 1/np.sum(M, axis=1)[:,np.newaxis]
    ew = 1/np.sum(M, axis=0)[np.newaxis,:]
    W = aw @ ew
    print('W.shape=', W.shape)
    WM = W*M
    simple_plot_3d(ax10, WM, x=q)

    WC = md.C*ew
    WP = WM @ np.linalg.pinv(WC)

    for c in WC:
        ax11.plot(c)
    plot_text(ax11, "$C_w$", fontsize=MATRIX_SYM_FONTSIZE)

    for p in WP.T:
        ax12.plot(q, p)
    plot_text(ax12, "$P_w$", fontsize=MATRIX_SYM_FONTSIZE)

    WP_ = WP/aw
    for p in WP_.T:
        ax22.plot(q, p)
    plot_text(ax22, "$P$", fontsize=MATRIX_SYM_FONTSIZE)

    fig.tight_layout()
    fig.subplots_adjust(top=0.92)
    plt.show()



"""

    StageGuinier.py

    Copyright (c) 2016-2024, SAXS Team, KEK-PF

"""
import os
import time
import numpy                as np
from openpyxl               import Workbook
from molass_legacy.KekLib.NumpyUtils import np_savetxt, simply_safe_sprintf
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
from molass_legacy.AutorgKek.KekToolsGP import AutorgKek, ErrorResult
from molass_legacy.AutorgKek.LightObjects import LightIntensity, LightResult
from molass_legacy.AutorgKek.AtsasTools import autorg as autorg_atsas
from molass_legacy.KekLib.ProgressInfo import put_info
from .ProgressInfoUtil import STREAM_GUINIER
from .AnalyzerUtil import make_guinier_result_stamp, exists_guinier_analysis_result
from molass_legacy.Reports.ReportUtils import get_header_record, make_record, make_guinier_analysis_book
from molass_legacy.GuinierAnalyzer.SimpleGuinier import QUALITY_BOUNDARY_QRG
from molass_legacy.GuinierAnalyzer.AutorgKekAdapter import AutorgKekAdapter
from molass_legacy.Test.Tester import compare_tester_rg
from molass_legacy.Reports.DefaultFont import set_default_font
set_default_font()

NUM_LATEST_RGS          = 5
LOG_DEFATED_MIN_STDEV   = 0.2
LOG_DEFATED_RATIO       = 2.0

def run_gunier_analysis(self, debug=False):
    print( 'run_gunier_analysis' )

    intensity_array_ = self.get_intensity_array()
    self.qvector = intensity_array_[0,:,0]

    self.guinier_result_array = []
    quality_array = []

    self.write_state( 'gunier_analysis-doing' )
    if self.using_avareged_files:
        datafiles_  = self.averaged_datafiles
    else:
        datafiles_  = self.datafiles

    fh = open( self.serial_file, 'w' )
    fh.write( get_header_record() + '\n' )
    num_successes = 0
    total_time = 0
    self.latest_rgs_atsas   = []
    self.latest_rgs_self    = []
    self.rg_ratio_range     = [ None, None ]
    self.head_gap_ratio_range   = [ None, None ]

    for i, datafile in enumerate( datafiles_ ):
        self.stop_check()

        self.current_datafile = datafile
        data = intensity_array_[ i, self.usable_slice, : ]
        # print( 'data.shape=', data.shape )
        try:
            t0 = time.time()
            if self.use_simpleguinier == 0:
                autorg_kek  = AutorgKek( data )
            else:
                autorg_kek  = AutorgKekAdapter( data )
            result_kek  = autorg_kek.run( robust=True, optimize=True )
            num_successes += 1
            total_time += time.time() - t0
        except Exception as exc:
            etb = ExceptionTracebacker()
            self.logger.error( etb )
            self.logger.error( 'current_datafile:' + str(self.current_datafile) )   # str to allow None
            result_kek = ErrorResult()

        result_atsas, result_atsas_eval = autorg_atsas( datafile )
        assert result_atsas is not None

        if result_atsas.Rg is not None and result_atsas_eval is None:
            self.logger.warning( 'unexpected None result_atsas_eval; error column value=0 may have caused this result' )

        compare_rg_stdev( self, [ result_kek.Rg, result_atsas.Rg ] )
        update_rg_ratio_range( self, result_kek )

        try:
            compare_tester_rg( i, result_kek.Rg, QUALITY_BOUNDARY_QRG )
        except:
            etb = ExceptionTracebacker()
            self.logger.error( etb )

        record = make_record( datafile, result_kek, result_atsas )
        fh.write( record + '\n' )
        fh.flush()

        # convert into light objects to facilitate garbage collection
        light_intensity = LightIntensity( autorg_kek.intensity )
        light_result    = LightResult( result_kek )

        self.guinier_result_array.append( [ light_intensity, light_result, result_atsas ] )
        quality = None if result_kek is None else result_kek.Quality

        if self.log_memory_usage == 1:
            m_info = self.process.memory_info()
            memory_usage    = " " + str( ( m_info.vms, ) )
        else:
            memory_usage    = ""

        _, file = os.path.split( datafile )

        if result_kek.Rg is None:
            time.sleep( 0.1 )   # to let the log window to handle this separately
            self.logger.warning( "Rg estimation failed in %s.%s" % ( file, memory_usage ) )
            time.sleep( 0.1 )   # to let the log window to handle this separately
        else:
            log_Rg      = simply_safe_sprintf( '%.3g', result_kek.Rg )
            log_quality = simply_safe_sprintf( '%.2g', quality )
            self.logger.info( "Rg from %s is estimated to be %s with quality %s.%s" % ( file, log_Rg, log_quality, memory_usage ) )

        if quality is None:
            quality = 0
        quality_array.append( quality )
        if debug:
            self.logger.info("put_info( (%d, 0), %d )", STREAM_GUINIER, i+1)
        put_info( (STREAM_GUINIER, 0), i+1 )
    fh.close()
    self.quality_array = np.array(quality_array)
    self.current_datafile = None

    self.logger.info( "average time for successful Rg evaluation was %.2g seconds" % ( total_time/num_successes ))
    self.logger.info( "rg_ratio_range=%s" % str( self.rg_ratio_range ) )
    self.logger.info( "head_gap_ratio_range=%s" % str( self.head_gap_ratio_range ) )
    self.write_state( 'gunier_analysis-done' )

    exists_guinier_analysis_result( self )  # to create folder
    save_guinier_result( self )

def compare_rg_stdev( self, rg_pair ):

    array_list = []
    for i, list_ in enumerate([self.latest_rgs_self, self.latest_rgs_atsas]):
        list_.append( rg_pair[i] )
        if len(list_) > NUM_LATEST_RGS:
            list_.pop(0)
        array_list.append( np.array( [ np.nan if v is None else v for v in list_ ] ) )

    if len( self.latest_rgs_self ) < 3:
        return

    stdev_k = np.std( array_list[0] )
    stdev_a = np.std( array_list[1] )

    if np.isnan(stdev_k) or np.isnan(stdev_a):
        return

    if False:
        if stdev_a > LOG_DEFATED_MIN_STDEV and stdev_k/stdev_a > LOG_DEFATED_RATIO:
            self.logger.warning( 'possibly defeated by atsas: sa stdev=%.3g, atsas stdev=%.3g' % ( stdev_k, stdev_a ) )

def update_rg_ratio_range( self, result_kek ):
    if result_kek.Rg is None:
        return

    if result_kek.rg_ratio_for_better_peak is None:
        return

    if self.rg_ratio_range[0] is None or result_kek.rg_ratio_for_better_peak < self.rg_ratio_range[0]:
        self.rg_ratio_range[0] = result_kek.rg_ratio_for_better_peak

    if self.rg_ratio_range[1] is None or result_kek.rg_ratio_for_better_peak > self.rg_ratio_range[1]:
        self.rg_ratio_range[1] = result_kek.rg_ratio_for_better_peak

    if self.head_gap_ratio_range[0] is None or result_kek.head_gap_ratio < self.head_gap_ratio_range[0]:
        self.head_gap_ratio_range[0] = result_kek.head_gap_ratio

    if self.head_gap_ratio_range[1] is None or result_kek.head_gap_ratio > self.head_gap_ratio_range[1]:
        self.head_gap_ratio_range[1] = result_kek.head_gap_ratio

def write_gunier_stamp( self ):
    fh = open( self.stamp_file, "w" )
    fh.write( make_guinier_result_stamp( self ) )
    fh.close()

def save_guinier_result( self ):
    if os.path.exists( self.guinier_file ):
        os.remove( self.guinier_file )
    os.rename( self.serial_file, self.guinier_file )
    write_gunier_stamp( self )
    np_savetxt( self.quality_file, self.quality_array )

def make_guinier_analysis_report( self ):
    # print( 'make_guinier_analysis_report' )

    if self.excel_is_available:
        wb = Workbook()
        ws = wb.active
    else:
        wb = self.result_wb
        ws = wb.create_sheet('Guinier Analysis')

    book_file = self.temp_folder + '/--serial_analysis-temp.xlsx'
    make_guinier_analysis_book( self.excel_is_available, wb, ws, book_file, self.guinier_file, self.c_vector,
                                    self.xr_j0,
                                    applied_ranges=self.applied_ranges,
                                    parent=self, logger=self.logger )
    self.temp_books.append( book_file )



"""
    V2PropOptimizer.V1RangeCoverter.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.BasicUtils import Struct       # as ControlInfoProxy
from molass_legacy.Decomposer.DecompUtils import make_range_info_impl
from molass_legacy.Decomposer.FitRecord import FitRecord
from molass_legacy.Decomposer.ModelEvaluator import ModelEvaluator
from molass_legacy.Decomposer.UnifiedDecompResult import UnifiedDecompResult
from RangeInfo import shift_editor_ranges
from molass_legacy.DataStructure.AnalysisRangeInfo import AnalysisRangeInfo, convert_to_paired_ranges

def convert_to_v1_model_ranges(prop_optimizer, peaks, paired_ranges, debug=False):
    x = prop_optimizer.x
    y = prop_optimizer.y
    model = prop_optimizer.model

    opt_recs = []
    for kno, params in enumerate(peaks):
        # see 
        sign = 1                # almost always, any exceptions?
        chisqr_n = np.nan       # may not be used
        accepts_real_x = True   # as a temporary fix in ExtrapolSolverDialog
        evaluator = ModelEvaluator(model, params, sign=sign, accepts_real_x=accepts_real_x)
        y_ = evaluator(x)
        m = np.argmax(y_)
        peak = Struct(top_x=x[m] - x[0])   # 
        fit_rec = FitRecord(kno, evaluator, chisqr_n, peak)
        opt_recs.append(fit_rec)

    unif_result = UnifiedDecompResult(x=x, y=y, opt_recs=opt_recs)
    control_info = unif_result.get_range_edit_info()
    editor_ranges = control_info.editor_ranges

    ret_ranges = make_range_info_impl(opt_recs, control_info)
    new_paired_ranges, num_ranges = convert_to_paired_ranges(ret_ranges)

    if debug:
        print("peaks.shape=", peaks.shape)
        print("peaks.paired_ranges=", paired_ranges)
        print("editor_ranges=", editor_ranges)
        print("ret_ranges=", ret_ranges)
        print("new_paired_ranges=", new_paired_ranges)

        shifted_ranges = shift_editor_ranges(x[0], editor_ranges)
        cy_list = prop_optimizer.compute_cy_list(peaks)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("convert_to_models_ranges")
            ax.plot(x, y)
            for cy in cy_list:
                ax.plot(x, cy, ":")
            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red")
            ymin, ymax = ax.get_ylim()
            for ranges in shifted_ranges:
                for f, t in ranges:
                    p = Rectangle(
                            (f, ymin),      # (x,y)
                            t - f,          # width
                            ymax - ymin,    # height
                            facecolor   = 'cyan',
                            alpha       = 0.2,
                        )
                    ax.add_patch(p)
            fig.tight_layout()
            plt.show()

    return new_paired_ranges, ret_ranges



# coding: utf-8
"""
    GeometryUtils.py

    Copyright (c) 2017-2022, Masatsuyo Takahashi, KEK-PF
"""
import numpy as np

def rotate( th, wx, wy ):
    cx  = wx[-1]
    cy  = wy[-1]
    wx_ = wx - cx
    wy_ = wy - cy
    c   = np.cos( th )
    s   = np.sin( th )
    return cx + ( wx_*c - wy_* s ), cy + ( wx_*s + wy_* c )

def rotated_argmin(theta, wy, debug=False):
    wy_ = wy/(np.max(wy) - np.min(wy))
    n = len(wy)
    wx_ = np.arange(n)/(n-1)
    rx, ry = rotate(theta, wx_, wy_)
    m = np.argmin(ry)
    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        print('m=', m)
        degrees = 180*theta/np.pi
        plt.push()
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14, 7))
        fig.suptitle("rotated_argmin debug", fontsize=20)
        ax1.set_title("Normalized Data", fontsize=16)
        ax2.set_title("Rotated %+.4g degrees" % degrees, fontsize=16)
        ax1.plot(wx_, wy_)
        ax1.plot(wx_[m], wy_[m], 'o', color='red')
        ax2.plot(rx, ry)
        ax2.plot(rx[m], ry[m], 'o', color='red')
        fig.tight_layout()
        fig.subplots_adjust(top=0.88)
        plt.show()
        plt.pop()
    return m

def rotated_argminmax(theta, wy, debug=False):
    wy_ = wy/(np.max(wy) - np.min(wy))
    n = len(wy)
    wx_ = np.arange(n)/(n-1)
    rx, ry = rotate(theta, wx_, wy_)
    m1 = np.argmin(ry)
    m2 = np.argmax(ry)

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        degrees = 180*theta/np.pi
        plt.push()
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14, 7))
        fig.suptitle("rotated_argmin debug", fontsize=20)
        ax1.set_title("Normalized Data", fontsize=16)
        ax2.set_title("Rotated %+.4g degrees" % degrees, fontsize=16)
        ax1.plot(wx_, wy_)
        ax2.plot(rx, ry)

        for m in [m1, m2]:
            print('m=', m)
            ax1.plot(wx_[m], wy_[m], 'o', color='red')
            ax2.plot(rx[m], ry[m], 'o', color='red')

        fig.tight_layout()
        fig.subplots_adjust(top=0.88)
        plt.show()
        plt.pop()

    return m1, m2

def polygon_area_centroid(polygon, num_division=10):
    path = polygon.get_path()
    xy = path.vertices
    xmin = np.min(xy[:,0])
    xmax = np.max(xy[:,0])
    ymin = np.min(xy[:,1])
    ymax = np.max(xy[:,1])
    x = np.linspace(xmin, xmax, num_division+1)
    y = np.linspace(ymin, ymax, num_division+1)
    x_ = (x[1]-x[0])/2 + x[:-1]
    y_ = (y[1]-y[0])/2 + y[:-1]
    xx, yy = np.meshgrid(x_, y_)
    points = np.array(list(zip(xx.flatten(), yy.flatten())))
    inside = path.contains_points(points)
    in_points = points[inside]
    return np.average(in_points, axis=0)



"""
    SecTheory.RetensionTime.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, root

def compute_retention_time(x, rgs):
    model_trs = []
    for rg in rgs:
        # x[0] : t0
        # x[1] : P
        # x[2] : rp
        # x[3] : m
        rho = 1 if rg > x[2] else rg/x[2]
        model_trs.append(x[0] + x[1]*(1 - rho)**x[3])
    return np.array(model_trs)

def compute_rg_from_retention_time(seccol_params, trs, init_rgs):
    t0, K, rp, m = seccol_params
    model_rg = []
    for i, tr in enumerate(trs):

        def fun(z):
            rho = 1 if z > rp else z/rp
            return t0 + K*(1 - rho)**m - tr

        sol = root(fun, [init_rgs[i]])
        model_rg.append(sol.x[0])

    return np.array(model_rg)

def make_initial_guess(trs, t0_upper_bound=0):
    Kinit = np.average(trs)*2
    bounds = ((-Kinit, +Kinit), (Kinit*0.2, Kinit*3), (50, 300), (1, 6))
    t0_init = t0_upper_bound - 30
    return (t0_init, Kinit, 80, 3), bounds

NEGATIVE_PENALTY_SCALE = 1e5
TZERO_PENALTY_SCALE = 1e5

def estimate_conformance_params(rgs, trs, rg_curve, t0_upper_bound=None, debug=False):

    if t0_upper_bound is None:
        from molass_legacy._MOLASS.SerialSettings import get_setting
        t0_upper_bound = get_setting("t0_upper_bound")

    min_rg = 0
    lim_tr = rg_curve.x[-1]
    zeros_rgs = np.zeros(len(rgs))

    def objective_func(z):
        # max_tr = compute_retention_time(z, [min_rg])        # i.e., max_tr = t0 + K
        max_tr = z[0] + z[1]    # i.e., max_tr = t0 + K
        t0_penalty = TZERO_PENALTY_SCALE * max(0, z[0] - t0_upper_bound)**2
        model_trs = compute_retention_time(z, rgs)
        max_tr_constraint = min(0, max_tr - lim_tr)**2      # so that max_tr > lim_tr
        return np.sqrt(np.average((model_trs - trs)**2)) + max_tr_constraint + t0_penalty

    initial_guess, bounds = make_initial_guess(trs, t0_upper_bound)

    # method = "Nelder-Mead"
    method = None
    result = minimize(objective_func, initial_guess, method=method, bounds=bounds)

    if debug:
        print("---------- len(rgs)=", len(rgs))
        print("---------- initial_guess=", initial_guess)
        print("---------- bounds=", bounds)
        print("---------- result.x=", result.x)

    return result

def estimate_conformance_params_fixed_poreexponent(rgs, trs, degree):

    def objective_func(x):
        x_ = np.concatenate([x, [degree]])
        model_trs = compute_retention_time(x_, rgs)
        return np.sqrt(np.average((model_trs - trs)**2))

    initial_guess, bounds = make_initial_guess(trs)

    # method = "Nelder-Mead"
    method = None
    result = minimize(objective_func, initial_guess[0:3], method=method, bounds=bounds[0:3])

    return result

def estimate_init_rgs(rg_curve, trs, t0_upper_bound, debug=False):
    x_, y_, rg_ = rg_curve.get_valid_curves()

    result = estimate_conformance_params(rg_, x_, rg_curve, t0_upper_bound)
    seccol_params = result.x

    mean = np.mean(rg_)
    std = np.std(rg_)
    init_rgs = np.linspace(mean+std, mean-std, len(trs))
    if debug:
        print("trs=", trs)
        print("init_rgs=", init_rgs)

    model_rgs = compute_rg_from_retention_time(seccol_params, trs, init_rgs)

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        from DataUtils import get_in_folder
        from scipy.interpolate import UnivariateSpline
        print("seccol_params=", seccol_params)

        in_foler = get_in_folder()
        x = rg_curve.x
        y = rg_curve.y
        t0 = seccol_params[0]
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("Initial Rg Estimation for %s" % in_foler, fontsize=16)
            axt = ax.twinx()
            axt.grid(False)
            if x is None:
                ax.plot(x_, y_)
            else:
                ax.plot(x, y)
            k = 0
            for xs, ys, rs in rg_curve.get_curve_segments():
                label = "observed rg" if k == 0 else None
                axt.plot(xs, rs, color="C1", label=label)
                k += 1
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            t = t0_upper_bound
            ax.plot([t, t], [ymin, ymax], ":", color="red", label="t0_upper_bound")

            spline = UnivariateSpline(x_, rg_, s=0, ext=3)
            xm = x[x > t0]
            rs = spline(xm)
            rm = compute_rg_from_retention_time(seccol_params, xm, rs)
            axt.plot(xm, rm, ":", color="green", label="estimated model rg curve")
            axt.plot(trs, model_rgs, "o", color="yellow", label="estimated initial rg's")

            ax.legend(loc="upper left")
            axt.legend(loc="upper right")
            fig.tight_layout()
            plt.show()

    return model_rgs, seccol_params



"""
    ErrorLogCheck.py

    Copyright (c) 2024, Masatsuyo Takahashi, KEK-PF
"""

# borrowed from https://stackoverflow.com/questions/4664850/how-to-find-all-occurrences-of-a-substring
def findall(p, s):
    '''Yields all the positions of
    the pattern p in the string s.'''
    i = s.find(p)
    while i != -1:
        yield i
        i = s.find(p, i+1)

def all_known_errors(buffer):
    error_count = 0
    for i in findall("ERROR", buffer):
        error_count += 1
    
    known_error_count= 0
    for i in findall("ERROR,root,No counter info", buffer):
        known_error_count += 1
    
    return error_count == known_error_count


"""

    LambertBeer.py

    Copyright (c) 2019-2025, SAXS Team, KEK-PF

"""
import logging
import copy
import numpy as np
from scipy.optimize import minimize
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy.DataStructure.SvdDenoise import get_denoised_data
from molass_legacy.KekLib.ThreeDimUtils import compute_plane
from molass_legacy._MOLASS.SerialSettings import get_setting
USE_PEAK_TOP_RIDGE = True
NEGATIVE_PENALTY = 1
POSITIVE_AREA_RATIO = 0.1

class BasePlane:
    def __init__(self, data, index, ecurve, denoise=False, rank=None, j0=0, debug=True):
        self.logger = logging.getLogger(__name__)

        if denoise:
            if rank is None:
                rank = len(ecurve.peak_info) + 1
            data = get_denoised_data(data, rank=rank)

        self.data = data
        self.index = index
        self.x = np.arange(data.shape[0])
        self.y = np.arange(data.shape[1])
        self.divided = data.shape[0] < 100
        self.NW = np.ones(data.shape)
        safe_level = ecurve.max_y*POSITIVE_AREA_RATIO
        self.NW[data < safe_level] = 0
        self.cd = get_setting('xr_bpa_option')
        self.consider_cd = self.cd == 2

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.DataStructure.MatrixData import simple_plot_3d, contour_plot

            print('data.shape=', data.shape, 'index=', index)
            print('type(ecurve)=', type(ecurve))

            with plt.Dp():
                fig = plt.figure(figsize=(21,6))
                ax1 = fig.add_subplot(131, projection='3d')
                ax2 = fig.add_subplot(132)
                ax3 = fig.add_subplot(133, projection='3d')

                ax1.set_title("Surface Plot", fontsize=16, y=1.08)
                simple_plot_3d(ax1, data)

                ax2.set_title("Contour Plot", fontsize=16)
                cp = contour_plot(ax2, data, levels=np.linspace(0.5, 10, 5)*safe_level)
                # fig.colorbar(cp, ax=ax2)  # TODO

                ax3.set_title("Weights for Negative Penalty", fontsize=16, y=1.08)
                simple_plot_3d(ax3, self.NW)

                fig.tight_layout()
                plt.show()

        self.ecurve = ecurve
        self.j0 = int(ecurve.x[0])
        # self.make_C_matrix()
        # self.make_P_matrix()

    def make_C_matrix(self):
        self.C = np.array([self.ecurve.sy])

    def make_P_matrix(self):
        ridge_list = []
        for info in self.ecurve.peak_info:
            top_j = info[1]
            if self.divided:
                sy = self.data[:,top_j]
            else:
                sy = smooth(self.data[:,top_j])
            ridge_list.append(sy)

        self.P = np.array(ridge_list).T

    def objective(self, params):
        if len(params) == 2:
            a = 0
            b, c = params
        else:
            a, b, c = params
        BP  = compute_plane(a, b, c, self.x, self.y)
        M_  = self.data - BP
        # ey_ = - (self.y*b + c)

        # print('peak_info=', self.ecurve.peak_info)

        P_list = []
        C_list = []
        start = 0
        for k, info in enumerate( self.ecurve.peak_info ):
            if k < len(self.ecurve.boundaries):
                stop = self.ecurve.boundaries[k] - self.j0
            else:
                stop = len(self.y)

            # print([k], start, stop)

            peak_j  = info[1] - self.j0
            P1 = M_[:,peak_j]
            P_list.append(P1)
            C_ = np.zeros(len(self.y))
            V_ = M_[self.index,start:stop]       # extract the corrected concentration
            """
            subtract baseline from V_ with a, b, c.
            must consider the scale between M_ and the SMP - standard mapping plane.
            investigate the objective function at this position.
            """
            # V_ = self.ecurve.sy[start:stop]
            C_[start:stop] = V_/np.max(V_)  # normalize to the max
            C_list.append(C_)
            if self.consider_cd:
                C_list.append(C_**2)

            start = stop

        C_ = np.array(C_list)
        if USE_PEAK_TOP_RIDGE and self.cd == 1:
            P_ = np.array(P_list).T
        else:
            P_ = M_ @ np.linalg.pinv(C_)
        PC = np.dot(P_, C_)
        N_ = copy.deepcopy(M_)
        N_[N_ > 0] = 0

        self.P = P_
        self.C = C_

        ret_val = np.linalg.norm((PC - M_)) + np.linalg.norm((self.NW*N_))*NEGATIVE_PENALTY

        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.DataStructure.ModeledData import simple_plot_3d
            print('params=', (a, b, c))
            print('ret_val=', ret_val)
            if self.divided:
                index = 0
                i_slice = slice(None, None)
                j_slice = slice(None, None)
                j_width = self.data.shape[1]
            else:
                index = self.index
                i_slice = slice(0,30)
                j_width = 50
                j_start = 180
                j_slice = slice(j_start,j_start+j_width)
            with plt.Dp():
                fig = plt.figure()
                ax = fig.add_subplot(111, projection='3d')
                ax.set_title(str(self.debug_info))

                x_ = np.ones(j_width)*index
                y_ = np.arange(j_width)
                for k in range(self.C.shape[0]):
                    z_ = C_[k,j_slice]
                    scale = self.P[0,k]
                    zm = z_*scale + (a*index + b*y_ + c)
                    ax.plot(x_, y_, zm, color='orange', linewidth=5)

                simple_plot_3d(ax, PC[i_slice,j_slice], alpha=0.2)
                simple_plot_3d(ax, M_[i_slice,j_slice], alpha=0.2)
                simple_plot_3d(ax, self.data[i_slice,j_slice], color='yellow', alpha=0.2)
                simple_plot_3d(ax, BP[i_slice,j_slice], color='red', alpha=0.2)

                fig.tight_layout()
                plt.show()

        return ret_val

    def solve(self, debug=False, debug_info=None):
        self.debug_info = debug_info
        data = self.data
        self.allow_nonzero_a = get_setting('allow_angular_slope_in_mf')

        bounds = []
        if self.allow_nonzero_a:
            bounds.append((-1e-2, +1e-2))
        bounds += [
            (-1e-2, +1e-2),
            (-1.0, +1.0),
        ]
        init_params = np.zeros(len(bounds))
        
        result  = minimize(self.objective, init_params, bounds=bounds)
        if len(init_params) == 2:
            a = 0
            b, c = result.x
        else:
            a, b, c = result.x
        self.params = (a, b, c)

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            print('self.params=', self.params)
            with plt.Dp():
                fig = plt.figure(figsize=(10,8))
                ax = fig.add_subplot(111, projection='3d')
                ax.set_title(str(debug_info))
                self.debug_plot(ax)
                fig.tight_layout()
                plt.show()

        self.logger.info('MF-baseplane solved with params: (%g, %g, %g)' % self.params)
        return self.params

    def debug_plot(self, ax, plot_plane=True, plane_color='red', view_init=None):
        data = self.data
        a, b, c = self.params

        i = self.x
        j = self.y

        print('params=', self.params)
        # print('len(i)=', len(i), 'P.shape=', self.P.shape, 'C.shape=', self.C.shape)

        xx, yy = np.meshgrid(i, j)
        zz = data[xx, yy]

        ax.plot_surface(xx, yy, zz, alpha=0.2)

        x_ = np.ones(data.shape[1])*self.index
        y_ = j
        for k in range(self.C.shape[0]):
            z_ = self.C[k,:]
            scale = self.P[0,k]
            zm = z_*scale + (a*self.index + b*y_ + c)
            ax.plot(x_, y_, zm, color='orange')

        for k, info in enumerate(self.ecurve.peak_info):
            top_j = info[1]
            y_ = np.ones(len(i))*top_j
            z_ = self.P[:,k]
            ax.plot(i, y_, z_, color='green')

        if plot_plane:
            x = i[[0,-1]]
            y = j[[0,-1]]
            xx, yy = np.meshgrid(x, y)
            zz = xx*a + yy*b + c
            ax.plot_surface(xx, yy, zz, color=plane_color, alpha=0.2)

            y_ = j
            z_ = self.index*a + y_*b + c
            ax.plot(x_, y_, z_, color=plane_color, linewidth=3)

        if view_init is not None:
            ax.view_init(*view_init)

    def get_params(self):
        return self.params

    def get_baseline(self, x):
        a, b, c = self.params
        return x*a + self.y*b + c

    def get_baseplane(self):
        a, b, c = self.params
        plane = compute_plane(a, b, c, self.x, self.y)
        return plane

def get_args(sd, pre_recog, debug=False):
    """
    it would be better if we could do without the following
    if branch.
    """
    if pre_recog is None:
        start = 0
        data = sd.intensity_array[:,:,1].T
    else:
        sd = pre_recog.get_pre_recog_copy()
        start, stop = pre_recog.get_default_angle_range(sd)

        data = sd.intensity_array[:,start:stop,1].T

    xray_slice = sd.xray_slice
    index = (xray_slice.start + xray_slice.stop)//2
    ecurve = sd.xray_curve

    index_ = index - start
    return data, index_, ecurve

def get_standard_baseline(sd, pre_recog, debug=False, logger=None):
    data, index_, ecurve = get_args(sd, pre_recog, debug)
    bp = BasePlane(data, index_, ecurve, debug=debug)
    bp.solve(debug=debug)

    if logger is not None:
        a, b, c = bp.get_params()
        c_ratio = c/ecurve.max_y
        logger.info('baseplane parames: a=%g, b=%g, c=%g, c_ratio=%g', a, b, c, c_ratio)

    return bp.get_baseline(index_)

def get_base_plane(data, index, debug=False):
    from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
    ey = data[index,:]
    ecurve = ElutionCurve(ey)

    bp = BasePlane(data, index, ecurve, debug=debug)
    bp.solve(debug)
    return bp

def compute_base_plane(data, index, ecurve, denoise=False):
    bp = BasePlane(data, index, ecurve, denoise=denoise)
    bp.solve()
    BP = bp.get_baseplane()
    return BP


"""
    Jupyter.Installer.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""

from molass_legacy.KekLib.RunPython import run_this_python

PACKAGES = ["jupyterlab", "ipympl"]
MAX_ITER = 80

def install_jupyterlab(parent=None):
    jupyterlab = PACKAGES[0]
    result = run_this_python("-m", "pip", "show", jupyterlab)
    if False:
        print("returncode=", result.returncode)
        print(result.stdout)
        print(result.stderr)
    if result.returncode == 0:
        return True

    import os
    import sys
    import psutil
    from molass_legacy.KekLib.TkCustomWidgets import MessageBox
    site_packages_path = os.path.join(os.path.dirname(sys.executable), "lib\\site-packages")

    hdd = psutil.disk_usage('/')

    GB = 2**30
    print ("Total: %d GiB" % (hdd.total/GB))
    print ("Used: %d GiB" % (hdd.used/GB))
    print ("Free: %d GiB" % (hdd.free/GB))

    yn = MessageBox.askyesno("Jupyter Lab Install Confirmation",
        "%s package in not installed in the following folder.\n"
        "    %s\n\n"
        "It may require up to 200MB (=0.2GB) depending on the state of package installation, "
        "while the current disc usage seems like\n\n"
        "    Total: %5.0f GB\n"
        "    Used : %5.0f GB\n"
        "    Free : %5.0f GB.\n\n"
        "Would you like to accept and procced?"
        % (jupyterlab, site_packages_path, hdd.total/GB, hdd.used/GB, hdd.free/GB),
        parent=parent)

    if not yn:
        return False

    if parent is None:
        for package in PACKAGES:
            print("installing", package)
            result = run_this_python("-m", "pip", "install", package)
            print("returncode=", result.returncode)
    else:
        from molass_legacy.KekLib.RunPython import run_this_python_stdout_lines
        from molass_legacy.KekLib.ProgressMinDialog import run_with_progress

        def install_jupyterlab(exe_queue):
            n = 0
            for i, package in enumerate(PACKAGES):
                print("installing", package)
                for j, line in enumerate(run_this_python_stdout_lines("-m", "pip", "install", package)):
                    print([j], line)
                    n += 1
                    exe_queue.put([0, n])
            exe_queue.put([1, n])
            return True

        def on_return(ret):
            print("--------------- on_return: ret=", ret)

        try:
            ret, error_info = run_with_progress(parent, install_jupyterlab, max_iter=MAX_ITER,
                                        title="Jupyter.Installer", on_return=on_return)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "run_with_progress failure: ")
            ret = None

    return True



"""
    UV.PlainCurve.py

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from sklearn.cluster import KMeans
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve

DIEHARD_LIMIT = 0.05        # 
FLATTENABLE_RATIO = 0.3     # > 0.21 for FER_OA_302, < 1.54 for 20211021
EXTINGUISHED_RATIO = 0.01   # 
CERTAINTY_LIMIT = 0.5
BRIGHT_LINE_WAVELENGTHS = [486, 581, 656]    # see https://www.klv.co.jp/corner/what-is-d2-lump.html
PEAK_CONSIDER_LIMIT = 0.5
MIN_NUM_CLUSTER_MEMBERS = 5

class PlainCurve(ElutionCurve):
    def __init__(self, x, y, guess_sigmoid=False, save_fig=False):
        self.x = x
        self.y = y
        self.sigmoid_params = None
        self.sigmoid_certainty = 0
        if guess_sigmoid:
            from molass_legacy.Trimming.Sigmoid import guess_bent_sigmoid
            ret = guess_bent_sigmoid(x, y, return_certainty=True, save_fig=save_fig)
            self.sigmoid_params = ret[0]
            self.sigmoid_certainty = ret[1]

    def save(self, path):
        np.savetxt(path, np.array([self.x, self.y]).T)

def check_diehardness(a_curve, a_curve2, debug=False):
    x = a_curve.x
    certainty = a_curve2.sigmoid_certainty
    if certainty < CERTAINTY_LIMIT:
        x0 = x[0]
    else:
        # good possibility of flow change
        x0 = a_curve2.sigmoid_params[1]
    y1 = a_curve.y
    y2 = a_curve2.y
    extinct_ratios = []
    for k, rec in enumerate(a_curve.peak_info):
        j = rec[1]
        if x[j] < x0:
            # skip a flow change (mis-recognized peak)
            pass
        else:
            ratio = y2[j]/y1[j]
            extinct_ratios.append((x[j], y2[j], ratio))
    if len(extinct_ratios):
        extinct_ratios = np.array(extinct_ratios)
        max_k = np.argmax(extinct_ratios[:,1])
        x_, y_, max_ratio = extinct_ratios[max_k]
        judge = y_ > y2[-1] and max_ratio > DIEHARD_LIMIT   # checking y_ > y2[-1] to avoid misjudge as in 20160628
    else:
        # as in 20200624_2
        judge, max_k, max_ratio = False, None, None

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("check_diehardness debug")
            ax1.plot(x, y1)
            ax2.plot(x, y2)
            if len(extinct_ratios) > 0:
                axt = ax2.twinx()
                axt.grid(False)
                axt.plot(*extinct_ratios.T, "o", color="red")
            fig.tight_layout()
            plt.show()

    return judge, max_k, max_ratio

def find_an_alternative_curve(D, wv, a_curve, index, logger, debug=False):
    ridge_curves = []

    pick = bisect_right(wv, get_setting('absorbance_picking'))
    zabs = get_setting('zero_absorbance')

    max_y = a_curve.max_y
    found = False
    i = None
    brights = []
    for w in BRIGHT_LINE_WAVELENGTHS:
        if w < wv[-1]:
            brights.append(bisect_right(wv, w))
    required_cond = wv > zabs
    for b in brights:
        required_cond[b-10:b+11] = False    # not near bright lines

    WSCALE = 0.0001

    for k, rec in enumerate(a_curve.peak_info):
        j = rec[1]
        y_pick = D[pick,j]
        if y_pick/max_y < PEAK_CONSIDER_LIMIT:
            # ignore minor peaks
            continue

        y = D[:,j]
        y -= np.min(y)
        extinguished = np.where(np.logical_and(required_cond, y/max_y < EXTINGUISHED_RATIO))[0]

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(get_in_folder())
                ax.plot(wv, y)
                ax.plot(wv[extinguished], y[extinguished], "o")
                fig.tight_layout()
                plt.show()

        if len(extinguished) > MIN_NUM_CLUSTER_MEMBERS*2:

            X = np.array([wv[extinguished]*WSCALE, y[extinguished]]).T
            n_clusters = min(10, len(extinguished)//MIN_NUM_CLUSTER_MEMBERS)
            kmeans = KMeans(n_clusters=n_clusters, random_state=0).fit(X)
            labels = kmeans.labels_
            scores = []
            wheres = []
            for k in range(n_clusters):
                where = labels == k
                x_ = extinguished[where]
                y_ = y[x_]
                m = np.mean(y_)
                s = np.std(y_)
                scores.append((m**2+s**2)/len(x_))
                wheres.append(where)
            k = np.argmin(scores)
            where = wheres[k]

            if debug:
                print("n_clusters=", n_clusters)
                scale = 10000
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title("KMeans Grouping debug")
                    ax.set_ylabel("Absorbance")
                    ax.set_xlabel("Wavelength")
                    ax.scatter(X[:,0]/WSCALE, X[:,1], c=labels, s=40, cmap='viridis')
                    ax.plot(X[where,0]/WSCALE, X[where,1], "o", color="red")
                    fig.tight_layout()
                    plt.show()

            where_ = np.where(where)[0]
            m = int(np.median(where_))

            found = True
            tail = extinguished[m:]
            i = tail[0]
            if debug:
                from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
                from time import sleep
                limit_y = max_y*EXTINGUISHED_RATIO
                debug_path = get_setting("debug_path")
                with plt.Dp():
                    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                    fig.suptitle("Finding an alternative Flat Curve for %s" % get_in_folder(), fontsize=20)
                    ax1.set_title("Whole Absorbance Curve", fontsize=16)
                    ax2.set_title("Zoomed Absorbance Curve", fontsize=16)
                    for ax in ax1, ax2:
                        ax.set_xlabel("Wavelength (nm)")
                        ax.set_ylabel("Absorbance")
                        ax.plot(wv, y, label="eno=%d" % j)
                        ax.plot(wv[extinguished], y[extinguished], "o", color="yellow", label="y/max_y < %g" % EXTINGUISHED_RATIO)
                        ax.plot(wv[tail], y[tail], "o", color="green", label="all-below tail")
                        ax.plot(wv[i], y[i], "o", color="red", label="first point of tail")
                    xmin, xmax = ax1.get_xlim()
                    ymin, ymax = ax1.get_ylim()
                    ax1.set_ylim(ymin, ymax)
                    w = wv[index]
                    ax1.plot([w, w], [ymin, ymax], ":", color="red", label="baseline wavelength")
                    xmin2 = xmin*0.1 + xmax*0.9
                    xmax2 = xmax
                    ymin2 = ymin
                    ymax2 = ymin*0.9 + ymax*0.1
                    ax2.set_xlim(xmin2, xmax2)
                    ax2.set_ylim(ymin2, ymax2)
                    ax2.plot([xmin2, xmax2], [limit_y, limit_y], ":", color="cyan", label="limit_y")
                    w = wv[i]
                    ax2.plot([w, w], [ymin, ymax], ":", color="red", label="baseline wavelength")
                    for ax in [ax1, ax2]:
                        ax.legend(fontsize=12)
                    fig.tight_layout()
                    if True:
                        plt.show()
                    else:
                        plt.show(block=False)
                        fig.savefig(debug_path)
                        sleep(1)

        elif len(extinguished) > 0:
            i = extinguished[0]

        if i is not None:
            ridge_curves.append((j, y, i))

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        with plt.Dp():
            fig, ax = plt.subplots()
            for j, y, i in ridge_curves:
                ax.plot(wv, y, label="j=%d" % j)
                ax.plot(wv[i], y[i], "o", color="red")
            ax.legend()
            fig.tight_layout()
            plt.show()

    ridge_curves = sorted(ridge_curves, key=lambda x: -x[2])

    if found:
        if i < index:
            logger.info("alternative curve found at wv[%d]=%g will be replaced by one with the specified (longer) wv[%d]=%g.", 
                        i, wv[i], index, wv[index])
            i = index
        x = a_curve.x
        y = D[i,:]
        ret_curve = PlainCurve(x, y)
        w = wv[i]
    else:
        ret_curve = None
        w = None

    return ret_curve, w

def flattenable(ratio):
    if ratio is None:
        # as in 20200624_2
        return True
    else:
        return ratio < FLATTENABLE_RATIO

def make_secondary_e_curve_at(D, wv, e_curve, logger, wavelen=None, debug=False, save_fig=False):
    if wavelen is None:
        wavelen = get_setting("zero_absorbance")

    index = min(D.shape[0]-1, bisect_right(wv, wavelen))
    y = D[index,:]
    try:
        curve = PlainCurve(e_curve.x, y, guess_sigmoid=True, save_fig=save_fig)
        diehard, max_k, ratio = check_diehardness(e_curve, curve, debug=debug)
        if diehard:
            curve_, w = find_an_alternative_curve(D, wv, e_curve, index, logger, debug=debug)
            if curve_ is None:
                x = e_curve.x

                if flattenable(ratio):
                    flat_y = y.copy()
                    info = e_curve.peak_info[max_k]
                    flat_y[info[0]:info[2]] = 0
                else:
                    # as in 20211021 where ratio == 1.54
                    flat_y = np.zeros(len(y))
                    logger.warning("flat_y for uv baseline guess has been set to all zeros due to ratio=%.3g", ratio)

                if debug:
                    import molass_legacy.KekLib.DebugPlot as plt
                    from molass_legacy.Elution.CurveUtils import simple_plot
                    with plt.Dp():
                        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                        fig.suptitle("Secondary Elutin Curve at %.3g nm" % wavelen, fontsize=20)
                        ax1.set_title("a_curve")
                        ax2.set_title("a_curve2")
                        simple_plot(ax1, e_curve)
                        ax2.plot(x, y)
                        ax2.plot(x, flat_y, ":")
                        fig.tight_layout()
                        plt.show()

                curve = PlainCurve(x, flat_y)
                logger.info("UV data seem to include a diehard peak like ferritin from the extinction ratio=%.3g", ratio)
            else:
                set_setting('flat_wavelength', w)
                logger.info("UV baseline will be estimated using a auto-corrected wavelength %.4g", w)
                curve = curve_
    except:
        etb = ExceptionTracebacker()
        logger.warning("failed to create curve at wavelength %.3g with exception %s" % (wavelen,  etb.last_lines()))
        curve = None    # but, will cause an exceotopn later

    # curve.save(r"D:\TODO\20230623\temp\secondary_e_curve-1.txt")

    return curve
